#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/20 10:15

from ..func.base import MyClass, ArgParseClass, func_done
from ..probe.miko.connect import BaseDevice
from ..func.mrun import MrunArgParse
from ..probe.snmp import Snmp


class NetDeviceClass(MyClass):
    def __init__(self, ip=None, _init_conn=True, *args, **kwargs):
        '''
        :param ip:
        :param miko_param:
        :param snmp_param:
        :param _init_conn: os可能需要通过查询得到，设置_init_conn以延迟
        :param args:
        :param kwargs:
        '''
        MyClass.__init__(self, *args, **kwargs)
        self.ip = ip
        if 'log' not in kwargs:
            kwargs['log'] = self.log
        self._init_conn = _init_conn
        if self._init_conn:
            self.init_conn()

    def init_conn(self):
        self.miko_param = self.kwargs.get('miko_param')
        self.snmp_param = self.kwargs.get('snmp_param')
        if isinstance(self.miko_param, dict):
            if self.ip:
                self.miko_param.update({'ip': self.ip})
            self.kwargs.update(self.miko_param)
            self.cli = BaseDevice(**self.kwargs)
            if not self.ip:
                self.ip = self.cli.ip
        if isinstance(self.snmp_param, dict):
            if self.ip:
                self.snmp_param.update({'ip': self.ip})
            self.kwargs.update(self.snmp_param)
            self.snmp = Snmp(**self.kwargs)
            if not self.ip:
                self.ip = self.snmp.ip
        return True, func_done()

    def do(self):
        '''
('comware', '10.163.105.55'),
('vrp', '10.163.49.81'),
('ios', '10.163.108.122'),
('nxos', '10.163.172.32'),
('dnos', '10.163.81.25'),
('ftos', '10.60.3.25'),
('ibmnos', '10.163.73.35'),
('powerconnect', '10.163.73.144'),
('iosxe', '10.80.172.13'),
('iosxr', '10.163.105.10'),
        :return:
        '''
        funcs = {'comware': self.do_comware,
                 'vrp': self.do_vrp,
                 'ios': self.do_ios,
                 'iosxe': self.do_iosxe,
                 'iosxr': self.do_iosxr,
                 'nxos': self.do_nxos,
                 'dnos': self.do_dnos,
                 'ftos': self.do_ftos,
                 'powerconnect': self.do_powerconnect,
                 'ibmnos': self.do_ibmnos,
                 'junos': self.do_junos,
                 }
        return funcs[self.cli.os]()

    def do_comware(self):
        return False, 'no code.'

    def do_vrp(self):
        return False, 'no code.'

    def do_ios(self):
        return False, 'no code.'

    def do_iosxe(self):
        return False, 'no code.'

    def do_iosxr(self):
        return False, 'no code.'

    def do_nxos(self):
        return False, 'no code.'

    def do_dnos(self):
        return False, 'no code.'

    def do_ftos(self):
        return False, 'no code.'

    def do_ibmnos(self):
        return False, 'no code.'

    def do_powerconnect(self):
        return False, 'no code.'

    def do_junos(self):
        # {'os': 'junos', 'ip': '95.3-1.124.252'}
        return False, 'no code.'


class NeteaseDeviceClass(NetDeviceClass):
    def __init__(self, *args, **kwargs):
        NetDeviceClass.__init__(self, *args, **kwargs)


class NetDeviceArgParse(ArgParseClass):
    def __init__(self, *args, **kwargs):
        ArgParseClass.__init__(self, *args, **kwargs)
        if not kwargs.get('delay'):
            self.add_device()

    def add_device(self, group='Single Device'):
        self.add('--ip', required=True, help='target IP address', group=group)
        self.add('--os', default='', help='OS of device', group=group)


class MultiNetDeviceArgParse(MrunArgParse):
    def __init__(self, *args, **kwargs):
        MrunArgParse.__init__(self, *args, **kwargs)
        if not kwargs.get('delay'):
            self.add_multi_device()

    def add_multi_device(self, group='Multi Device'):
        self.add('--limit', type=int, default=100000, help='query limit', group=group)
        self.add('--ips', default='', help='spec ips', group=group)
        self.add('--os', default='', help='os of single jobs, do NOT check libre', group=group)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/14 11:50

import hashlib
import time
from functools import partial

import requests

from ..func.log import make_logger


def im(content, url, auth, mode, to, add_time=False, log=False):
    '''
    :param s:
    :param url:
    :param auth:
    :param mode:
    :param to:
    :param add_time:    服务端自动添加时间
    :param log:
    :return:
    '''
    # clock = datetime_to_string(now())
    # _content = '{0} \n{1}'.format(clock, content)
    payload = {'timestamp': int(time.time()),
               'mode': mode,
               'auth': auth,
               'to': to,
               'content': content,
               'add_time': add_time,
               }
    payload['sign'] = hashlib.md5('{timestamp}{mode}{auth}{to}{content}'.format(**payload).encode('utf-8')).hexdigest()
    if log:
        make_logger(console=False, write=True).info('{}'.format(content))
    return requests.post(url, payload)


popo = partial(im, mode='popo')
yx = partial(im, mode='yx')
sms163 = partial(im, mode='sms163')
sms = partial(im, mode='sms')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/13 10:02

import re
from functools import wraps

from netmiko import Netmiko

from . import IbmnosTelnet
from ..func.base import MyClass, catch_exception, count_time
from ..func.string import replace


class BaseDevice(MyClass):
    def __init__(self, ip, os, passwords, *args, **kwargs):
        '''
        :param args:
        :param kwargs:
            ip
            port:   optional, default 22
            os
            passwords   [{'username':, 'password':, 'secret':}, {}, ...]
            timeout:    optional, default 3
            blocking_timeout
            device_type
                specific device_type for cli
            log:    optional
            verbose:    optional, default False
            ssh_enable: enable telnet login, default True
            telnet_enable: enable telnet login, default False
        '''
        MyClass.__init__(self, *args, **kwargs)
        self.ip = ip
        self.os = os
        self.passwords = [passwords] if isinstance(passwords, dict) else passwords
        self.device_type = kwargs.get('device_type')
        self.ssh_enable = kwargs.get('ssh_enable', True)
        self.telnet_enable = kwargs.get('telnet_enable')

    @catch_exception(ignore=True)
    def login(self, auth):
        '''
        :param auth:
            username, password, secret
        对于dell_dnos9， 有时会报 'Oops, unhandled type' 错误，
        根据 https://github.com/ktbyers/netmiko/issues/373，设置 use_keys=True, allow_agent=True，运行正常
        '''
        # paramiko.transport may except 'Oops, unhandled type' warning. (dell_dnos9 on 10.60.5.16)
        # https://github.com/ktbyers/netmiko/issues/373
        if auth['device_type'] == 'ibmnos_telnet':
            net_connect = IbmnosTelnet(**auth)
        else:
            # timeout, blocking_timeout
            if re.search('dnos', auth['device_type']) and 'use_keys' not in auth:
                extra = dict(use_keys=True, allow_agent=True)
            else:
                extra = {}
            net_connect = Netmiko(**auth, **extra, timeout=self.kwargs.get('timeout', 20),
                                  blocking_timeout=self.kwargs.get('cli_blocking_timeout', 20),
                                  )
        self.session = net_connect
        self.cache['cli_protocol'] = auth['device_type']
        if not self.device_type:
            self.device_type = auth['device_type']
        return True, 'connected.'

    def set_device_types(self):
        if self.device_type:
            device_types = self.device_type if isinstance(self.device_type, list) else [
                self.device_type]
        else:
            device_types = []
            if self.os in ['ftos', 'dnos']:
                if self.ssh_enable:
                    device_types.append('dell_dnos9')
                if self.telnet_enable:
                    device_types.append('dell_dnos6_telnet')
            elif self.os in ['comware']:
                if self.ssh_enable:
                    device_types.append('hp_comware')
                if self.telnet_enable:
                    device_types.append('hp_comware_telnet')
            elif self.os in ['nxos']:
                if self.ssh_enable:
                    device_types.append('cisco_nxos')
                if self.telnet_enable:
                    device_types.append('cisco_ios_telnet')
            elif self.os in ['iosxr']:
                if self.ssh_enable:
                    device_types.append('cisco_xr')
                if self.telnet_enable:
                    device_types.append('cisco_xr_telnet')
            elif self.os in ['iosxe']:
                if self.ssh_enable:
                    device_types.append('cisco_xr')
                if self.telnet_enable:
                    device_types.append('cisco_xr_telnet')
            elif self.os in ['ios']:
                if self.ssh_enable:
                    device_types.append('cisco_ios')
                if self.telnet_enable:
                    device_types.append('cisco_ios_telnet')
            elif self.os in ['vrp']:
                if self.ssh_enable:
                    device_types.append('huawei_vrpv8')
                if self.telnet_enable:
                    device_types.append('huawei_telnet')
            elif self.os in ['powerconnect']:
                if self.ssh_enable:
                    device_types.append('dell_powerconnect')
                if self.telnet_enable:
                    device_types.append('dell_powerconnect_telnet')
            elif self.os in ['ibmnos']:
                if self.telnet_enable:
                    device_types.append('ibmnos_telnet')
            elif self.os in ['junos']:
                if self.ssh_enable:
                    device_types.append('juniper_junos')
                if self.telnet_enable:
                    device_types.append('juniper_junos_telnet')
        return device_types

    @catch_exception()
    @count_time
    def get_session(self):
        auth = {'host': self.ip}
        if not self.cache.get('cli_false'):
            if not hasattr(self, 'session'):
                device_types = self.set_device_types()
                if not device_types:
                    return False, 'no device type for {} cli.'.format(self.os)
                for device_type in device_types[:]:
                    for pw in self.passwords[:]:
                        auth.update(pw)
                        auth['device_type'] = device_type
                        is_ok, result = self.login(auth)
                        self.cache['login_result'] = result
                        if is_ok:
                            if self.os in ['ibmnos']:
                                self.send_command('terminal-length 0')
                            elif self.os in ['powerconnect']:
                                self.send_command('terminal length 0')
                            self.cache['auth'] = auth
                            return is_ok, result
                        else:
                            if re.search('(Authentication|login)', result, re.I):
                                continue
                            if re.search(r'(time[d -]*out)', result, re.I):
                                break
            else:
                return True, 'cli connection exist.'

        self.cache['cli_false'] = True
        return False, 'login false.'

    # @staticmethod
    def check_cli(self, f):
        def wrapper(*args, **kwargs):
            is_ok, connection = self.get_session()
            if is_ok:
                return f(*args, **kwargs)
            return False, self.cache.get('login_result') or 'no cli connection'

        return wrapper

    @catch_exception()
    def send_command(self, commands, show_cmd=False, one_line_cmd=False, **kwargs):
        @self.check_cli
        def wrap(commands):
            if isinstance(commands, str):
                commands = commands.split('\n')
            output = []
            if one_line_cmd:
                commands = ['\n'.join(commands)]
            for cmd in commands:
                if not cmd.strip():
                    continue
                if show_cmd:
                    output.append('>>>>> {} <<<<<\n'.format(cmd))
                output.append('{}'.format(self.session.send_command(cmd, **kwargs)))
            return True, '\n'.join(output)

        return wrap(commands)

    @catch_exception()
    def display(self, commands, func=None, func_kw=None, tidy=True, show_cmd=True, one_line_cmd=False, **kwargs):
        '''

        :param commands:
        :param func: 回调结果处理函数
        :param func_kw:
        :param tidy: 删除重复的换行
        :param show_cmd: 结果中显示命令
        :param one_line_cmd: 不拆分命令
        :return:
        '''
        is_ok, result = self.send_command(commands, show_cmd=show_cmd, one_line_cmd=one_line_cmd, **kwargs)
        assert is_ok, result
        if tidy:
            result = replace(result)
        if func:
            if not func_kw:
                func_kw = {}
            return is_ok, func(result, **func_kw)
        return is_ok, result

    @catch_exception()
    def enable(self, *args, **kwargs):
        '''# 进入enable模式, netmiko 2.3.3中，base_connection.py在调用check_enable_mode时没有带参数，始终返回True，不能用
        '''

        @self.check_cli
        def wrap(*args, **kwargs):
            return self.session.enable(*args, **kwargs)

        return True, wrap(*args, **kwargs)


def try_login(device_types, passwords):
    def deco(f):
        @wraps(f)
        def wrapper(inst, auth):
            is_ok, result = False, 'init'
            for device_type in device_types:
                for pw in passwords:
                    auth.update(pw)
                    auth['device_type'] = device_type
                    is_ok, result = f(inst, auth)
                    # break when success or timeout. Except 'broken pipe' when auth fail on ftos.
                    if is_ok or (isinstance(result, str) and re.search('(timed out|connection refused)', result, re.I)):
                        break
                if is_ok:
                    break
            if is_ok:
                inst.cache['protocol'] = device_type
            return is_ok, result

        return wrapper

    return deco



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/6/11 15:54

import re
import time

from netmiko.base_connection import BaseConnection
from netmiko.ssh_exception import NetMikoAuthenticationException


class IbmnosTelnet(BaseConnection):
    def __init__(self, **kwargs):
        kwargs['device_type'] = 'ibmnos_telnet'
        BaseConnection.__init__(self, **kwargs)

    def telnet_login(
            self,
            pri_prompt_terminator=r"#\s*$",
            alt_prompt_terminator=r">\s*$",
            username_pattern=r"(?:user:|username|login|user name)",
            pwd_pattern=r"assword",
            delay_factor=1,
            max_loops=20,
    ):
        """Telnet login. Can be username/password or just password."""
        delay_factor = self.select_delay_factor(delay_factor)
        time.sleep(1 * delay_factor)

        output = ""
        return_msg = ""
        i = 1
        while i <= max_loops:
            try:
                output = self.read_channel()
                return_msg += output

                # Search for username pattern / send username
                if re.search(pwd_pattern, output, flags=re.I):
                    self.write_channel(self.password + self.TELNET_RETURN)
                    time.sleep(1 * delay_factor)
                    output = self.read_channel()
                    return_msg += output

                # Search for password pattern / send password
                if re.search('Select Command', output, flags=re.I):
                    self.write_channel('iscli' + self.TELNET_RETURN)
                    time.sleep(0.5 * delay_factor)
                    output = self.read_channel()
                    return_msg += output
                    if re.search(
                            pri_prompt_terminator, output, flags=re.M
                    ) or re.search(alt_prompt_terminator, output, flags=re.M):
                        return return_msg

                # Support direct telnet through terminal server
                if re.search(r"initial configuration dialog\? \[yes/no\]: ", output):
                    self.write_channel("no" + self.TELNET_RETURN)
                    time.sleep(0.5 * delay_factor)
                    count = 0
                    while count < 15:
                        output = self.read_channel()
                        return_msg += output
                        if re.search(r"ress RETURN to get started", output):
                            output = ""
                            break
                        time.sleep(2 * delay_factor)
                        count += 1

                # Check for device with no password configured
                if re.search(r"assword required, but none set", output):
                    self.remote_conn.close()
                    msg = "Login failed - Password required, but none set: {}".format(
                        self.host
                    )
                    raise NetMikoAuthenticationException(msg)

                # Check if proper data received
                if re.search(pri_prompt_terminator, output, flags=re.M) or re.search(
                        alt_prompt_terminator, output, flags=re.M
                ):
                    return return_msg

                self.write_channel(self.TELNET_RETURN)
                time.sleep(0.5 * delay_factor)
                i += 1
            except EOFError:
                self.remote_conn.close()
                msg = "Login failed: {}".format(self.host)
                raise NetMikoAuthenticationException(msg)

        # Last try to see if we already logged in
        self.write_channel(self.TELNET_RETURN)
        time.sleep(0.5 * delay_factor)
        output = self.read_channel()
        return_msg += output
        if re.search(pri_prompt_terminator, output, flags=re.M) or re.search(
                alt_prompt_terminator, output, flags=re.M
        ):
            return return_msg

        self.remote_conn.close()
        msg = "Login failed: {}".format(self.host)
        raise NetMikoAuthenticationException(msg)


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/12 10:46

import os
import re
import shutil
import socket
import traceback
from pathlib import Path

import psutil

TERM_WIDTH, TERM_DEPTH = shutil.get_terminal_size((80, 20))
LIB_DIR = 'seakylib'


def get_hostname():
    return socket.gethostname()


def get_if(name=None):
    d = psutil.net_if_addrs()
    if name:
        return {k: v for k, v in d.items() if k.startswith(name)}
    return d


def get_pwd():
    '''
    即使被上层调用，也会返回lib所在目录
    '''
    r = Path(os.path.split(os.path.realpath(__file__))[0])
    parts = r._parts
    return os.path.join(*parts[:parts.index(LIB_DIR)])


def get_caller(top=True):
    '''返回用户调用函数，过滤lib'''
    es = traceback.extract_stack()
    l = es if top else es[::-1]
    for x in l:
        if not re.search('({}|pydev)'.format(LIB_DIR), x.filename):
            return Path(x.filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/4/20 14:44

import queue
import time
import traceback
from multiprocessing import Process, Manager
from random import randint

from ..func.base import ArgParseClass
from ..func.base import MyClass
from ..net.db import MyModel


class MultiRun(MyClass):
    def __init__(self, func, func_kws, process_num=5, inline=False, func_common_kw=None, func_common_kw_obj=None,
                 **kwargs):
        '''
        :param func: 处理函数
        :param func_kws:    需要处理的参数
        :param process_num: 进程数
        :param inline: 单进程运行
        :param func_common_kw: 函数通用参数, 不能放入log object
        :param func_common_kw_obj: 函数通用参数, 可放入object，如log，因为object不能被dump，所以单独处理
        :return
        '''
        MyClass.__init__(self, **kwargs)
        self.func = func
        self.func_kws = func_kws
        self.process_num = process_num
        self.func_common_kw = {} if func_common_kw is None else func_common_kw
        self.func_common_kw_obj = {} if func_common_kw_obj is None else func_common_kw_obj
        self.inline = inline
        m = Manager()
        self.q_input = m.Queue()
        self.q_output = m.Queue()

    def job(self, process_i):
        while True:
            try:
                v = self.q_input.get_nowait()
                is_ok, result = self.func(**v['kw'], **self.func_common_kw_obj)
                v.update({'is_ok': is_ok, 'result': result})
                self.q_output.put(v)
            except queue.Empty:
                break
            except Exception as e:
                self.log.info(traceback.format_exc())
                break

    def run(self):
        if not isinstance(self.func_kws, list):
            return False, 'func_kws is not list.'
        idxes = []
        for i, kw in enumerate(self.func_kws, 1):
            idx = id(kw)
            kw.update(self.func_common_kw)
            d = {'order_in': i, 'idx': idx, 'kw': kw}
            idxes.append(idx)
            self.q_input.put(d)

        if self.inline:
            self.job(1)
        else:
            process_num = min(self.process_num, len(self.func_kws))
            ps = []
            for process_i in range(1, process_num + 1):
                process = Process(target=self.job, args=([process_i]))
                ps.append(process)
                process.start()
            for i, p in enumerate(ps):
                p.join()

        outputs = {}
        for i in range(1, self.q_output.qsize() + 1):
            d = self.q_output.get_nowait()
            d['order_out'] = i
            outputs[d['idx']] = d
        return True, [outputs.get(idx, {'is_ok': False, 'message': 'result is not exist.'}) for i, idx in
                      enumerate(idxes)]


def update_results(model, datas, key, sql=None, cls_conn_str=None, session=None, last_cols=None, timed=False):
    '''
    :param model: 结果表格
    :param datas: 运行结果
    :param key: 主键
    :param sql: 如果无sql，则返回表格数据
    :param cls_conn_str: db连接字串
    :param session:     db session
    :param last_cols:   [col1, col2], 更新时需要保留的上一次状态。model中需要有 col1_last, col2_last字段
    :param timed:   记录时间
    :return:
    '''
    if model.__class__ == MyModel:
        mm = model
    else:
        if cls_conn_str:
            MyModel.cls_conn_str = cls_conn_str
            mm = MyModel(model)
        elif session:
            mm = MyModel(model, session=session)
        else:
            return
    if not last_cols:
        last_cols = [col.replace('_last', '') for col in mm.cols_name if col.endswith('_last')]
    if sql:
        is_ok, data_old = mm.query(sql=sql, key=key)
    else:
        is_ok, data_old = mm.query(key=key)
    for i, v in enumerate(datas):
        k = key(v) if hasattr(key, '__call__') else v[key]
        if k in data_old:
            _keys = list(v.keys())
            for col in _keys:
                if col not in mm.cols_name:
                    continue
                if col in last_cols:
                    v['{}_last'.format(col)] = data_old[k][col]
        count_fail_col = 'failed'
        if 'is_ok' in v and count_fail_col in mm.cols_name:
            if k in data_old:
                _count = data_old[k][count_fail_col]
                if not _count:
                    _count = 0
            else:
                _count = 0
            if not v['is_ok']:
                v[count_fail_col] = _count + 1
            else:
                v[count_fail_col] = _count
        if 'is_ok' in v:
            v['is_ok'] = 1 if v['is_ok'] else 0

    is_ok, result = mm.update(data_new=datas, data_old=data_old, key=key, timed=timed)
    return is_ok, result


class MrunArgParse(ArgParseClass):
    def __init__(self, *args, **kwargs):
        ArgParseClass.__init__(self, *args, **kwargs)
        ArgParseClass.add_base(self)
        if not kwargs.get('delay'):
            self.add_multi()

    def add_multi(self, group='Multi Process'):
        self.add('--process_num', type=int, default=100, help='multi process number', group=group)
        self.add('--inline', action='store_true', default=False, help='inline mode', group=group)


if __name__ == '__main__':
    def test(i):
        # must return is_ok, message
        t = randint(0, 3)
        time.sleep(t)
        msg = 'i am {}, waiting {}.'.format(i, t)
        print(msg)
        return True, msg


    mr = MultiRun(func=test, func_kws=[{'i': x} for x in range(5)], process_num=3, inline=False)
    print(mr.run())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/9/4 18:22

'''
    modify ~/.pyenv/versions/3.x.x/lib/python3.x/site-packages/puresnmp/transport.py
    ~/.pyenv/versions/3.7.2/lib/python3.7/site-packages/puresnmp/transport.py
        BUFFER_SIZE = 4096 # 4 KiB -> 20480
    ~/.pyenv/versions/3.x.x/lib/python3.x/site-packages/puresnmp/api/pythonic.py
        line 115, func multiwalk(), timeout=2 -> 10
'''

from collections import OrderedDict, defaultdict

from puresnmp import bulkwalk, bulkget
from puresnmp.x690.util import tablify

from ..func.base import MyClass
from ..func.string import bytes_decode, replace

MIB = {
    'general':
        {
            'sysName': '1.3.6.1.2.1.1.5.0',
            'sysDescr': '1.3.6.1.2.1.1.1.0',
            'ifNumber': '1.3.6.1.2.1.2.1.0',
            'dot1dBasePortIfIndex': '1.3.6.1.2.1.17.1.4.1.2',
            'ifName': '1.3.6.1.2.1.31.1.1.1.1',
            'ifDescr': '1.3.6.1.2.1.2.2.1.2',
            'ipNetToMediaPhysAddress': '1.3.6.1.2.1.4.22.1.2',  # arp table
            'dot1qTpFdbPort': '1.3.6.1.2.1.17.7.1.2.2.1.2',  # mac-address table
            'lldpRemPortId': '1.0.8802.1.1.2.1.4.1.1.7',
            'lldpRemPortDesc': '1.0.8802.1.1.2.1.4.1.1.8',
            'lldpRemSysName': '1.0.8802.1.1.2.1.4.1.1.9',
            'lldpRemSysDesc': '1.0.8802.1.1.2.1.4.1.1.10',
        },
    'h3c': {'dot3adAggPortSelectedAggID': '1.2.840.10006.300.43.1.2.1.1.12',
            'dot3adAggPortAttachedAggID': '1.2.840.10006.300.43.1.2.1.1.13',
            'hh3cAggPortListSelectedPorts': '1.3.6.1.4.1.25506.8.25.1.1.1.6'}}


class Snmp(MyClass):
    def __init__(self, ip, community, port=161, *args, **kwargs):
        MyClass.__init__(self, *args, **kwargs)
        self.ip = ip
        self.community = community
        self.port = port
        self.snmp_auth = {'ip': ip, 'community': community, 'port': port}

    def lookup_num_oid(self, name):
        '''
        查找oid，可以被重写
        :param name:
        :return:
        '''
        if 'mib' not in self.cache:
            self.cache['mib'] = {}
            for k, v in MIB.items():
                self.cache['mib'].update(v)
        return self.cache['mib'].get(name, name)

    def bulkget(self, *args, **kwargs):
        '''
        :param args:
        :param kwargs:
        :return:
        '''
        for x in ['scalar_oids', 'repeating_oids']:
            if kwargs.get(x):
                # 把scalar_oids, repeating_oids转成 {'x.x.x.x.x': name}, 方便返回
                if isinstance(kwargs[x], dict):
                    continue
                oids = [kwargs[x]] if isinstance(kwargs[x], str) else kwargs[x]
                oids = {self.lookup_num_oid(y): y for y in oids}
                if x == 'scalar_oids':
                    # scalar oid要去掉.0
                    oids = {(k[:-2] if k.endswith('.0') else k): v for k, v in oids.items()}
                kwargs[x] = oids
        return self._bulkget(*args, **kwargs)

    def _bulkget(self, scalar_oids=None, repeating_oids=None, max_list_size=1000, timeout=5, keys=None, ret_raw=False):
        '''
        scalar_oids, repeating_oids 必须是 dict
        :param scalar_oids:
            {'1.3.6.1.2.1.1.1': 'name1', '1.3.6.1.2.1.2.1': 'name2'} -> {'name1': value1, 'name2': value2}
        :param repeating_oids:  会返回OrderedDict, 有max_list_size限制, 建议用bulkwalk
        :param max_list_size:
        :param timeout:
        :param keys: 替换snmp oid中的item为字符串，优先使用oids dict中的value
            {'1': 'xxx', '2': 'yyy'}
        :param rat_raw:  返回原始的数据
        :return:
        '''
        raw = bulkget(scalar_oids=scalar_oids.keys() if isinstance(scalar_oids, dict) else scalar_oids,
                      repeating_oids=repeating_oids.keys() if isinstance(repeating_oids, dict) else repeating_oids,
                      max_list_size=max_list_size,
                      timeout=timeout, **self.snmp_auth)
        d = raw._asdict()
        for k, v in d.items():
            for k1, v1 in v.items():
                v[k1] = bytes_decode(v1)
        if ret_raw:
            return d
        d1 = defaultdict(OrderedDict)
        if not keys:
            keys = {}
        for k, v in d.items():
            if k == 'scalars' and scalar_oids:
                for k1, v1 in v.items():
                    name = scalar_oids.get(k1) or scalar_oids.get(k1[:-2]) or k1
                    name = keys.get(name, name)
                    d1[k][name] = v1
            elif repeating_oids:
                # 比较乱，待整理，用bulkwalk先
                # oid中的.在replace函数中容易被误匹配, 需要escape
                pats = [(x + '.', '') for x in repeating_oids]
                for k1, v1 in v.items():
                    index, oid = replace(k1, pats=pats, default='', ret_with_pat=True, _any=True, escape=True)
                    name = repeating_oids.get(oid[:-1]) or oid[:-1]
                    key = keys.get(name, name)
                    if index not in d1[k]:
                        d1[k][index] = {}
                    d1[k][index].update({key: v1})
        return d1

    def bulkwalk(self, oids, *args, **kwargs):
        if isinstance(oids, str):
            oids = [oids]
        oids = {self.lookup_num_oid(y): y for y in oids}
        return self._bulkwalk(oids=oids, *args, **kwargs)

    def _bulkwalk(self, oids, keys=None, index=None, bulk_size=1000, num_base_nodes=0, ret_raw=False, orderby=None,
                  to_str=True, proc=None):
        '''
        :param oids: dict
        :param keys: 替换snmp oid为字符串
            {'1': 'xxx', '2': 'yyy'}
        :param index:   数字, 返回dict, 以index为key, 会忽略空值
        :param bulk_size:
        :param num_base_nodes:
        :param ret_raw: 返回迭代器, 因为有时需要详细的oid, 但tablify返回不带oid
            for row in raw:
                print('%s: %r' % row)
        :param orderby: 排序
        :param to_str: 返回str，否则可能转成int或float
        :param proc: (func, *args), 不使用tablify，proc, 返回(key, value)组成字典
        :return:
        '''
        raw = bulkwalk(oids=oids.keys(), bulk_size=bulk_size, **self.snmp_auth)
        if proc:
            func, *args = proc
            d = OrderedDict()
            for x in raw:
                k, v = func(x, *args)
                d[k] = v
            return d
        if ret_raw:
            # [(k, bytes_decode(v, to_type=to_type)) for k, v in raw]
            return raw
        lst = tablify(raw, num_base_nodes=num_base_nodes)
        if orderby:
            if isinstance(orderby, str):
                lst.sort(key=lambda x: x[orderby])
            elif hasattr(orderby, '__call__'):
                lst.sort(key=orderby)
        if keys:
            if '0' not in keys:
                keys['0'] = 'id'
            _lst = []
            for i, x in enumerate(lst):
                d = {keys.get(k, k): bytes_decode(x[k], to_type='str' if to_str else None) for k, v in x.items()}
                _lst.append(d)
            if not index or index not in keys:
                return _lst
            _dict = OrderedDict()
            _index = keys[index]
            for i, x in enumerate(_lst):
                if not x[_index]:
                    continue
                _dict[x[_index]] = x
            return _dict
        else:
            return [{k: bytes_decode(v, to_type='str' if to_str else None) for k, v in x.items()} for i, x in
                    enumerate(lst)]


if __name__ == '__main__':
    sn = Snmp('192.168.2.63', 'xxxxxxxx')
    ret = sn.bulkget(scalar_oids=['1.3.6.1.2.1.1.1', '1.3.6.1.2.1.2.1'],
                     repeating_oids=['1.3.6.1.2.1.31.1.1.1.1', '1.3.6.1.2.1.31.1.1.1.18'],
                     keys={'1.3.6.1.2.1.1.1': 'descr', '1.3.6.1.2.1.2.1': 'port',
                           '1.3.6.1.2.1.31.1.1.1.1': 'ifName', '1.3.6.1.2.1.31.1.1.1.18': 'ifAlias'})
    sn.bulkwalk(oids=['1.3.6.1.2.1.47.1.1.1.1.2', '1.3.6.1.2.1.47.1.1.1.1.5'],
                keys={'2': 'descr', '5': 'clas', '7': 'name', '11': 'sn', '13': 'model'},
                index='2', orderby='2')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/6 14:49

import hashlib
import smtplib
from copy import deepcopy
from email import encoders
from email.header import Header
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path

from ..func.base import MyClass, run_func
from ..func.string import str_is_number, comma_digit


class Mail(MyClass):
    '''
    使用本地服务器时，设置 /etc/hosts为FQDN，否则远端会拒绝连接。
    127.0.0.1   xxxxx.cn xxxxx
    很多服务器也会拒绝虚拟的FQDN
    运行 hostname -f
    sudo dpkg-reconfigure exim4-config
    '''

    def __init__(self, server='localhost', port=25, username=None, password=None, conn_mode='plain', **kwargs):
        MyClass.__init__(self, **kwargs)
        self.cache['error'] = []
        self.smtp = None
        self.server, self.port, self.username, self.password, self.conn_mode = server, port, username, password, conn_mode
        self.login()

    def login(self):
        try:
            if self.server == 'localhost':
                smtp = smtplib.SMTP()
                smtp.connect(self.server, self.port)
            else:
                if self.conn_mode == 'tls':
                    # tls加密方式，通信过程加密，邮件数据安全，使用正常的smtp端口
                    smtp = smtplib.SMTP(self.server, self.port)
                    smtp.set_debuglevel(True)
                    smtp.ehlo()
                    smtp.starttls()
                    smtp.ehlo()
                    smtp.login(self.username, self.password)
                elif self.conn_mode == 'ssl':
                    # 纯粹的ssl加密方式，通信过程加密，邮件数据安全, smtps 465
                    smtp = smtplib.SMTP_SSL(self.server, self.port)
                    smtp.ehlo()
                    smtp.login(self.username, self.password)
                else:
                    smtp = smtplib.SMTP()
                    smtp.connect(self.server, self.port)
                    smtp.login(self.username, self.password)
            self.smtp = smtp
            return True
        except Exception as e:
            self.log_error(e)

    def check_path(self, fn):
        if Path(fn).exists():
            return True
        else:
            self.cache['error'].append('{} is not exist.'.format(fn))

    def send(self, from_addr=None, to_addrs=None, **kwargs):
        '''公共server会验证from_addr'''
        from_addr = str(Header(from_addr or self.from_addr, 'utf-8'))
        to_addrs = to_addrs or self.to_addrs
        if self.cc_addrs:
            to_addrs = to_addrs + self.cc_addrs
        if self.bcc_addrs:
            to_addrs = to_addrs + self.bcc_addrs
        if not self.cache.get('error'):
            try:
                r = self.smtp.sendmail(from_addr=from_addr, to_addrs=to_addrs, msg=self.msg.as_string(), **kwargs)
                return True
            except Exception as e:
                pass
        else:
            print('error occur.')

    def create_message(self, from_addr, to_addrs, subject, cc_addrs=None, bcc_addrs=None, charset='utf-8'):
        '''to_addrs, from_addr for display in body'''
        if isinstance(to_addrs, str):
            to_addrs = [to_addrs]
        if isinstance(cc_addrs, str):
            cc_addrs = [cc_addrs]
        if isinstance(bcc_addrs, str):
            bcc_addrs = [bcc_addrs]
        self.from_addr = from_addr
        self.to_addrs = to_addrs
        self.cc_addrs = cc_addrs
        self.bcc_addrs = bcc_addrs
        msg = MIMEMultipart()
        # msg['From'] = Header(from_addr, charset=charset)
        msg['From'] = from_addr
        msg['To'] = ', '.join(to_addrs)
        msg['Subject'] = Header(subject, charset=charset)
        if cc_addrs:
            msg['Cc'] = ', '.join(cc_addrs)
        self.msg = msg

    def add_text(self, s, _subtype='html', _charset='utf-8'):
        self.msg.attach(MIMEText(s, _subtype, _charset))

    def add_image(self, fn, cid):
        '''如果要在正文中显示图片，需要在在正文中的html中引用 <img src="cid:XXXXX">, 然后为图片添加 Content-ID: <XXXXX>, 有<>。
        filename与content-id没有关系，任意，不写也行
<html lang="utf-8">
    <body>
        <div>
            <div><table border="1"><tbody>
                <tr><td><b>row1</b></td><td>{}</td></tr>
                <tr><td><b>row2</b></td><td>{}</td></tr>
            </tbody></table></div>
            <div><a href="{graph_url}"><img src="cid:{}"></a></div>
            <br>
        </div><br>
        <div>Generated by: 10.220.16.43:~/project/cacti/Report.py</div>
    </body>
</html>
        '''
        if self.check_path(fn):
            mime = MIMEBase('image', 'png', filename=Path(fn).name)
            mime.add_header('Content-Disposition', 'inline', filename=Path(fn).name)
            mime.add_header('Content-ID', '<{0}>'.format(cid))
            mime.add_header('X-Attachment-Id', '{0}'.format(cid))
            mime.set_payload(open(fn, 'rb').read())
            encoders.encode_base64(mime)
            self.msg.attach(mime)
            return True

    def add_attach(self, fn=None, content=None, disp_name=''):
        if fn:
            if self.check_path(fn):
                att = MIMEApplication(open(fn, 'rb').read())
                att.add_header('Content-Disposition', 'attachment', filename=disp_name or Path(fn).name)

        elif content:
            att = MIMEApplication(content)
            att.add_header('Content-Disposition', 'attachment',
                           filename=disp_name or '{}.att'.format(hashlib.md5(content).hexdigit()))
        else:
            self.cache['error'].append('no attach found.')
            return
        self.msg.attach(att)
        return True

    def quit(self):
        if self.smtp:
            self.smtp.close()


def html_row_alternated_bg(row, i=0):
    row['tr_css'].update({'background-color': ['#FFFFFF', '#F0F0F0'][i % 2]})


def html_row_focus(row, col=None, thres=None, color='#FF0000'):
    if col not in row:
        return
    if str_is_number(row[col]):
        # 强制转float
        value = float(row[col])
        if isinstance(thres, (int, float)):
            if value > thres:
                row['tr_css'].update({'color': color})
        else:
            if value > row[col] > thres[0]:
                row['tr_css'].update({'color': color})


def html_row_to_tr(row, keys=None, css=None, comma=False):
    if not css:
        css = {}
    if isinstance(row, list):
        s_css = (' style="{}"'.format(';'.join(['{}: {}'.format(k, v) for k, v in css.items()])))
        s = '<tr{}>{}</tr>'.format(s_css, ''.join(['<td>{}</td>'.format(x) for x in row]))
    elif isinstance(row, dict):
        s_attr = (' ' + ';'.join(['{}="{}"'.format(k, v) for k, v in row['tr_attr'].items()]))
        css.update(row.get('tr_css') or {})
        s_css = (' style="{}"'.format(';'.join(['{}: {}'.format(k, v) for k, v in css.items()])))
        s = '<tr{}{}>{}</tr>'.format(s_attr, s_css, ''.join(
            ['<td>{}</td>'.format(comma_digit(row[k]) if comma and 'id' not in k else row[k]) for k in keys]
        ))
    else:
        s = ''
    return s


def html_df_to_table(df, titles=None, common_css=None, makeup=None, table_style='', comma=False):
    '''
    在df设置tr_attr, tr_css可以设置表格tr的属性
    :param df:    dict or dataframe
    :param titles:  [col1, col2, ...] or [(col1, disp1), (col2, disp2), ...], 可裁剪df
    :param common_css:  dict
    :param table_style:   "margin: auto;"
    :param makeup:    关注某字段
    :param comma:    逗号分割数字
    :return:
    '''

    html = list()
    html.append('''<div style="font-family:微软雅黑,Verdana,&quot;Microsoft Yahei&quot;,SimSun,sans-serif;font-size:14px; line-height:1.6;">
<table border=1 style="border: 1px groove black; table-layout: fixed;{}"><tbody>'''.format(table_style))
    if not titles:
        titles = [(x, x) for x in df.columns.to_list()]
    elif len(titles[0]) == 1:
        titles = [(x, x) for x in titles]

    if not common_css:
        common_css = {'text-align': 'center'}

    css_header = deepcopy(common_css)
    css_header.update({'background-color': '#FFFFC0', 'font-weight': 'bold'})
    html.append(html_row_to_tr(row=[x[1] for x in titles], css=css_header))

    if not makeup:
        makeup = []
    elif not isinstance(makeup, list):
        makeup = [makeup]

    for i, row in enumerate(df.to_dict('records')):
        if not row.get('tr_attr'):
            row['tr_attr'] = {}
        if not row.get('tr_css'):
            row['tr_css'] = {}
        row['tr_css'].update(common_css)
        html_row_alternated_bg(row, i)
        for func in makeup:
            run_func(func, row=row)
        html.append(html_row_to_tr(row, keys=[x[0] for x in titles], comma=comma))

    html.append('</tbody></table></div>')
    ret = '\n'.join(html)
    return ret


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/28 10:04

import numpy as np
import pandas as pd

from .string import change_type


def set_none(df, other=None, **kwargs):
    '''将nan替换为None, 可以应用于series'''
    return df.where(pd.notnull(df), other=other, **kwargs)


def drop_nan(df, col):
    '''按某一列nan舍弃行'''
    return df[pd.notnull(df[col])]


def select_row(df, func):
    '''
    :param df:
    :param func: return True/False
    :return:
    '''
    df1 = df.apply(func, axis=1)
    return df1[df1]


def to_df(data, cols=None, black=None, sort_by=None, sort_force_number=False, sort_desc=False, index_disp=False,
          index_name=None, use_none=True, **kwargs):
    '''
    :param data    dict/list/df
    :param cols:   []
    :param black:  cols黑名单
    :param sort_by:    str/dict, 以字段排序
    :param sort_force_number:    sort_key转数字排序
    :param sort_desc:    降序
    :param index_disp:    显示index
    :param index_name:    index名字，index_name-> df.index.name-> 'index'
    :param use_none:   替换nan为None
    :param kwargs:
        orient: columns/index   dict的key
        columns: 如果orient为index，可以分配columns
    '''
    df = pd.DataFrame.from_dict(data, **kwargs) if isinstance(data, (dict, list)) else data
    if index_disp:
        index_name = index_name or df.index.name or 'index'
        df[index_name] = df.index
        df = df[list(df)[-1:] + list(df)[:-1]]
    if not isinstance(black, list):
        black = []
    if not isinstance(cols, list):
        cols = df.columns.tolist()
    df = df[[x for x in cols if x in df.columns.tolist() and x not in black]]
    if sort_by:
        origin_columns = df.columns.tolist()
        if isinstance(sort_by, str):
            sort_by = [sort_by]
        if sort_force_number:
            _sort_by = []
            for co in sort_by:
                if df[co].dtype == np.object:
                    # 强制转np.object为float
                    key = '_sort_{}'.format(co)
                    try:
                        df[key] = df[co].astype(float, errors='raise')
                    except Exception:
                        # 混合情况
                        df[key] = df[co].apply(change_type, to_type=float, default=0)
                    _sort_by.append(key)
                else:
                    _sort_by.append(co)
        else:
            _sort_by = sort_by
        df = df.sort_values(by=_sort_by, axis=0, ascending=not sort_desc, inplace=False)[origin_columns]
    if use_none:
        df = set_none(df)
    return df


def array_to_df(lst, index=None, columns=None, **kwargs):
    if not index:
        index = range(int(len(lst) / len(columns)))
    if not columns:
        columns = range(int(len(lst) / len(index)))
    return pd.DataFrame(np.array(lst).reshape(len(index), len(columns)), index=index, columns=columns, **kwargs)


def array_to_np(list, **kwargs):
    return np.array(list, **kwargs)


def add_columns(df, col, value, axis=1):
    '''
    :param df:
    :param col: str/list
    :param value:
        如果value是函数，如果如果col是list，则value需要是一个函数，返回list；如果col是str, value返回str。
        如果没有匹配，最好返回np.nan，可以后续dropna()
    :return:
    '''
    if hasattr(value, '__call__'):
        if isinstance(col, list):
            def func(*args, **kwargs):
                return pd.Series(value(*args, **kwargs))

            df[col] = df.apply(func, axis=axis)
        else:
            df[col] = df.apply(value, axis=axis)
    elif isinstance(value, dict):
        df[col] = pd.Series(value)
    elif isinstance(value, list):
        df[col] = value
    else:
        df[col] = value


def df_dump(df, path, mode='csv', **kwargs):
    if mode == 'pickle':
        df.to_pickle(path, **kwargs)
    elif mode == 'csv':
        # index=False
        df.to_csv(path, **kwargs)


def df_load(path, mode='csv', **kwargs):
    if mode == 'pickle':
        return pd.read_pickle(path, **kwargs)
    elif mode == 'csv':
        return pd.read_csv(path, **kwargs)


def change_df_type():
    '''
    1、pd.to_numeric(series)
    2、df['x'].astype(str)
    3、df = df.infer_objects()'''
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/14 11:56

from datetime import datetime

# pip3 install python-dateutil
from dateutil.relativedelta import relativedelta


def datetime_now():
    return datetime.now()


def last_month():
    return datetime.now() - relativedelta(months=1)


def datetime_to_string(dt=None, fmt="%Y-%m-%d %H:%M:%S"):
    if not dt:
        dt = datetime_now()
    return dt.strftime(fmt)


def string_to_datetime(s, fmt):
    return datetime.strptime(s, fmt)


def datetime_to_timestamp(dt=None):
    if not dt:
        dt = datetime.now()
    return int(datetime.timestamp(dt))


def timestamp_to_datetime(timestamp):
    return datetime.fromtimestamp(timestamp)


def timestamp_to_string(timestamp, *args, **kwargs):
    return datetime_to_string(timestamp_to_datetime(timestamp), *args, **kwargs)


def string_to_timestamp(s, fmt):
    return int(datetime_to_timestamp(string_to_datetime(s, fmt)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/20 14:56

import ipaddress
import re

import IPy


def is_ip(ip):
    if not ip:
        return False
    p = re.compile('^((25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(25[0-5]|2[0-4]\d|[01]?\d\d?)$')
    if p.match(ip):
        return True
    else:
        return False


def in_network(ip, network):
    nw = IPy.IP(network) if isinstance(network, str) else network
    return ip in nw


def in_network1(ip, network):
    nw = ipaddress.ip_network(network) if isinstance(network, str) else network
    return ipaddress.ip_address(ip) in nw


ip2int = lambda x: sum(int(y) * 256 ** i for i, y in enumerate(x.split('.')[::-1]))


def make_long_mask(short_mask):
    if not short_mask:
        return '255.255.255.0'
    s = ('1' * (int(short_mask) if short_mask else 24)).zfill(32)[::-1]
    return '.'.join([str(int(s[i * 8:i * 8 + 8], 2)) for i in range(4)])


def make_short_mask(long_mask):
    return 24 if not long_mask \
        else long_mask if isinstance(long_mask, int) \
        else len(''.join(bin(int(x))[2:].replace('0', '') for x in long_mask.split('.')))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2018/9/17 15:56


import re
import traceback
from collections import OrderedDict
from functools import wraps

import pandas as pd
from sqlalchemy import Float
from sqlalchemy import INTEGER as INTEGER1, FLOAT, create_engine, UniqueConstraint
from sqlalchemy.dialects.mysql import BIGINT, INTEGER, SMALLINT, TINYINT
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy import event
from sqlalchemy import exc
import os

from ..func.pd import set_none
from ..func.string import replace
from ..func.time import datetime_to_string


# sqlacodegen --outfile=librenms.py mysql+pymysql://test:test@xx.xx.xx.xx:3306/testdb --tables <table>


def catch_sql_exception(f):
    @wraps(f)
    def wrap(*args, **kwargs):
        inst = args[0]
        try:
            return f(*args, **kwargs)
        except AssertionError as e:
            return False, e
        except Exception as e:
            inst.session.rollback()
            return False, '{0}, {1}, {2}, {3}'.format(traceback.format_exc() if getattr(inst, 'verbose') else e,
                                                      f.__name__, args, kwargs)

    return wrap


def check_writable(f):
    @wraps(f)
    def wrap(self, *args, **kwargs):
        if hasattr(self.model, 'readonly') and self.model.readonly:
            return False, 'readonly on {}'.format(self.model.__tablename__)
        return f(self, *args, **kwargs)

    return wrap


def obj2dict(obj, key=None, cols=None):
    if isinstance(obj, dict) or not obj:
        return obj
    elif isinstance(obj, list):
        d = {}
        l = []
        if hasattr(obj[0], '__table__'):
            cols = [co.name for co in obj[0].__table__.columns]
        for i, ob in enumerate(obj):
            d1 = obj2dict(ob, cols=cols)
            if key in cols:
                d[d1[key]] = d1
            else:
                l.append(d1)
        return d if key in cols else l
    elif hasattr(obj, '__table__'):
        if not cols:
            cols = [co.name for co in obj.__table__.columns]
        return {co: getattr(obj, co) for co in cols}


def add_quote(v, to_str=True, split=False, strip=True, quote='"'):
    '''
    添加 " " 到sql语句中
    :param v: int/float/list/str
    :param to_str:
    :param split: 如果split, 则分割v, 返回合并, xxx,yyy -> "xxx","yyy"
    :param strip: 去除空值
    :return:
    '''
    if v is None:
        return 'Null'
    elif isinstance(v, list):
        return [add_quote(x, to_str=to_str, split=False, strip=strip, quote=quote) for x in v]
    elif isinstance(v, (int, float)):
        return '{0}{1}{0}'.format(quote, v) if to_str else str(v)
    elif split:
        return ','.join(add_quote([x for x in v.split(',') if (not strip or x.strip())],
                                  to_str=to_str, split=False, strip=strip, quote=quote))
    else:
        return '{0}{1}{0}'.format(quote, replace(v, pats=['(^"|"$)', "(^'|'$)"]))


class MyModel:
    cls_conn_str = ''
    _engine = None

    def __init__(self, model=None, **kwargs):
        '''
        :param model:   Model of table
        :param session: db_session
        :param engine_kw: kwargs of self.engine
        :param kwargs:
            adapt_length:   更改时适应字符串长度
        '''
        self.cache = {'quote_cols': {}}
        self.kwargs = kwargs
        self.model = model
        if model:
            self.init_model()
        self.pk = None
        self.verbose = kwargs.get('verbose')

    def init_model(self):
        self.cols = {x.name: x for x in self.model.__table__.columns}
        self.cols_unicon = [(col for col in x.columns) for x in self.model.__table_args__ if
                            isinstance(x, UniqueConstraint)] if hasattr(self.model, '__table_args__') else []
        self.cols_name, self.cols_notnullable, self.cols_unique = list(), list(), list()
        for name, co in self.cols.items():
            self.cols_name.append(name)
            if not co.nullable:
                self.cols_notnullable.append(co.name)
            if co.unique or co.primary_key:
                self.cols_unique.append(co.name)
            if co.primary_key:
                self.pk = co.name
        # last_cols for update last_<>
        self.cache['last_cols'] = []
        for co in self.cols_name:
            last_co = 'last_{}'.format(co)
            if last_co in self.cols_name:
                self.cache['last_cols'].append(co)


    @classmethod
    def engine(cls, conn_str=None, echo=False, real=False, **kwargs):
        '''
        :param conn_str:
        :param echo:
        :param real:
            if real, new session with new engine will create a real new session, REAL is REQUIRED in multi process mode
            otherwise, func assigns the exist _engine, which may reuse session existed
        :param kwargs:
        :return:
        '''
        if not real and cls._engine:
            return cls._engine
        if not conn_str:
            conn_str = cls.cls_conn_str
        cls._engine = create_engine(conn_str, echo=echo, pool_pre_ping=True, pool_recycle=120, **kwargs)

        @event.listens_for(cls._engine, "connect")
        def connect(dbapi_connection, connection_record):
            connection_record.info['pid'] = os.getpid()

        @event.listens_for(cls._engine, "checkout")
        def checkout(dbapi_connection, connection_record, connection_proxy):
            pid = os.getpid()
            if connection_record.info['pid'] != pid:
                connection_record.connection = connection_proxy.connection = None
                raise exc.DisconnectionError(
                    "Connection record belongs to pid %s, "
                    "attempting to check out in pid %s" %
                    (connection_record.info['pid'], pid)
                )
        return cls._engine

    @classmethod
    def new_session(cls, real=False):
        if real:
            return Session(bind=cls.engine(real=real))
        _Session = sessionmaker(bind=cls.engine())
        return _Session()

    @property
    def session(self):
        if self.kwargs.get('session'):
            return self.kwargs['session']
        elif self.cache.get('session'):
            return self.cache['session']
        else:
            self.cache['session'] = self.new_session(**self.kwargs.get('engine_kw', {}))
            return self.cache['session']

    def add_quote_by_col_type(self, value, name):
        '''
        判断是否要在sql中加""
        :param value:
        :param name:
        :return:
        '''
        if value is None:
            return 'null'
        if not name in self.cache['quote_cols']:
            types = (Float, FLOAT, BIGINT, TINYINT, SMALLINT, INTEGER, INTEGER1)
            no_quote = isinstance(self.cols[name].type, types)
            self.cache['quote_cols'][name] = no_quote
        else:
            no_quote = self.cache['quote_cols'][name]
        if isinstance(value, str):
            value = value.replace('"', '\\"')
            if self.kwargs.get('adapt_length') and hasattr(self.cols[name].type, 'length'):
                value = value[:self.cols[name].type.length]
        return ('{}' if no_quote else '"{}"').format(value)

    @catch_sql_exception
    def get_obj(self, data=None, key='id', condition='', retdict=False):
        '''
        :param data:    dict of data
        :param key:
        :param condition:
        :return:
        '''

        def make_condition(col, v):
            return '{0}={1}'.format(col.name, v if col.type in [INTEGER, FLOAT] else "'{}'".format(v))

        if data:
            if key in data:
                result = self.session.query(self.model).filter(
                    getattr(self.model, key) == data[key]).first()
                if retdict:
                    result = obj2dict(result)
                return True, result

            for k, v in data.items():
                if k in self.cols_unique:
                    result = self.session.query(self.model).filter(getattr(self.model, k) == v).first()
                    if retdict:
                        result = obj2dict(result)
                    return True, result
            for x in self.cols_unicon:
                for col in x:
                    if col.name not in data.keys():
                        continue
                condition = ' and '.join([make_condition(y, data[y.name]) for y in x])
        if not condition:
            return False, 'no condition provided on get_obj.'
        sql = 'select * from {0} where {1}'.format(self.model.__tablename__, condition)
        result = self.session.execute(sql).first()
        return True, result

    @check_writable
    @catch_sql_exception
    def update_one(self, data, ex_obj=None, key='id', expandable=False):
        '''
        使用self.update代替
        :param data:
        :param ex_obj:
        :param key:
        :param expandable:
        :return:
        '''
        if ex_obj is None:
            status, ex_obj = self.get_obj(data)
            if not status:
                return status, ex_obj
        if not ex_obj:
            if expandable:
                return self.add(data, expandable=False, check=False)
            else:
                return False, '{} is not exist.'.format(getattr(ex_obj, key))
        flag_change = False
        if isinstance(ex_obj, dict):
            if key not in ex_obj:
                return False, 'key {0} is not in model when update {1}'.format(key, ex_obj)
            changes = []
            for k, v in data.items():
                if k in ex_obj and k in self.cols_name and v != ex_obj[k]:
                    changes.append('`{0}`={1}'.format(k, add_quote(v)))
                    flag_change = True
            if changes:
                sql = 'update `{0}` set {1} where `{2}`={3}'.format(self.model.__tablename__, ','.join(changes),
                                                                    key, add_quote(ex_obj[key]))
                self.session.execute(sql)
                # self.flush()
                self.commit()
            if not data.get(key):
                data[key] = ex_obj[key]
        else:
            for k, v in data.items():
                if k in self.cols_name and v != getattr(ex_obj, k):
                    flag_change = True
                    setattr(ex_obj, k, v)
            if flag_change:
                self.session.commit()
            if not data.get(key):
                data[key] = getattr(ex_obj, key)
        return True, {key: data[key], 'changed': flag_change}

    # def update(self, data_new, *args, **kwargs):
    #     # return self.update_df(data_new, *args, **kwargs)
    #     return self.update_batch(data_new, *args, **kwargs)

    def _update_compare_dict(self, data_new, data_old, key, cols=None, func_skip=None, last=True):
        if isinstance(data_new, pd.DataFrame):
            if data_new.index.name != key:
                data_new.index = data_new[key]
            data_new = data_new.to_dict(orient='index')
        if isinstance(data_old, pd.DataFrame):
            if data_old.index.name != key:
                data_old.index = data_old[key]
            data_old = data_old.to_dict(orient='index')
        items_new = []
        items_need_update = []
        changed_cols = []  # Align the changed keys
        items_skip = []
        only_change_last = 0

        for k, v in data_new.items():
            if hasattr(func_skip, '__call__') and func_skip(v):
                items_skip.append(v)
                continue
            if k not in data_old:
                items_new.append(v)
                continue
            if last:
                # 更新last_col
                for c in self.cache['last_cols']:
                    if c in data_old[k].keys() and 'last_{}'.format(c) not in v:
                        v['last_{}'.format(c)] = data_old[k][c]
            changed_real = False
            changed_last = False
            for k1, v1 in v.items():
                if cols and k1 not in cols:
                    continue
                if k1 in data_old[k] and k1 in self.cols_name and v1 != data_old[k][k1]:
                    if k1 not in changed_cols:
                        changed_cols.append(k1)
                    if k1.startswith('last_'):
                        changed_last = True
                    else:
                        changed_real = True
            if not changed_real and not changed_last:
                continue
            if changed_last and not changed_real:
                only_change_last += 1
            items_need_update.append((v, data_old[k]))
        items_miss = []
        for k, v in data_old.items():
            if k not in data_new:
                items_miss.append(v)
        return {'data_new': data_new, 'data_old': data_old,
                'items_new': items_new, 'items_need_update': items_need_update, 'only_change_last': only_change_last,
                'items_miss': items_miss, 'items_skip': items_skip,
                'changed_cols': changed_cols}

    def _update_compare_df(self, data_new, data_old, key, cols, func_skip, to_str):
        if isinstance(data_new, dict):
            data_new = pd.DataFrame(data_new).T
        if isinstance(data_old, dict):
            data_old = pd.DataFrame(data_old).T

        df_new_copy = data_new.sort_values(by=key, ascending=True, inplace=False)
        df_old_copy = data_old.sort_values(by=key, ascending=True, inplace=False)

        if to_str:
            df_new_copy = df_new_copy.astype(str)
            df_old_copy = df_old_copy.astype(str)
        for i, x in enumerate([df_new_copy, df_old_copy]):
            if x.index.name != key:
                x.index = x[key]
                # 把key作为index，需要修改index.name
                x.index.name = '{}_{}'.format(x.index.name, i)
        df1 = df_new_copy[[key]].merge(df_old_copy[[key]], indicator=True, how='outer', on=key)
        # df1 = data_new[[key]].merge(data_old[[key]], indicator=True, how='outer', on=key)
        compare = {n: g[key].tolist() for n, g in df1.groupby('_merge')}
        if cols:
            common_columns = [cols] if isinstance(cols, str) else cols
        else:
            common_columns = list(set(df_new_copy.keys()).intersection(set(df_old_copy.keys())))
        common_index = compare['both']
        # 相同的index和columns才能比较
        df_diff = df_new_copy[common_columns].loc[common_index] != df_old_copy[common_columns].loc[common_index]
        df_diff = df_diff[df_diff].dropna(axis=1, how='all', inplace=False).dropna(axis=0, how='all', inplace=False)

        if func_skip and hasattr(func_skip, '__call__'):
            df_diff1 = df_diff.apply(func_skip, 1)
            list_diff = df_diff1[df_diff1 == False].index.tolist()
            items_skip = df_new_copy.loc[df_diff1[df_diff1 == True].index.tolist()].to_dict(orient='index')
        else:
            list_diff = df_diff.index.tolist()
            items_skip = []

        changed_cols = df_diff.keys().tolist()
        items_need_update = []
        for x in list_diff:
            items_need_update.append((df_new_copy.loc[x].to_dict(), df_old_copy.loc[x].to_dict()))
        items_new = df_new_copy.loc[compare['left_only']]
        items_new = set_none(items_new).to_dict(orient='index')
        items_miss = df_new_copy.loc[compare['right_only']].to_dict(orient='index')

        # items_new = df_new_copy.loc[df_diff.index].to_dict(orient='index')

        return {'data_new': data_new, 'data_old': data_old,
                'items_new': items_new, 'items_need_update': items_need_update,
                'items_miss': items_miss, 'items_skip': items_skip,
                'changed_cols': changed_cols}

    @check_writable
    @catch_sql_exception
    def update(self, data_new, data_old=None, key=None, cols=None, force_condition=None,
               action_add=True, action_delete=False, action_mark_miss=False, func_skip=None,
               param_query=None, param_add=None, param_update=None, param_delete=None,
               with_detail=False, with_sqls=False,
               timed=True, last=True,
               use_df=False, df_to_str=False,
               block=2000, dryrun=False, **kwargs):
        '''
        :param data_new:    dict or list
        :param data_old:    dict or condition
        :param key: 需要一个unique的列作为key, 这个col也将作为query和delete的key
        :param cols: 指定比较某列, 如果为None, 自动发现
        :param force_condition: 强制条件，key may not the real unique, which is only valid in some spec condition.
            'ifName = xxxx and device = yyyy', extra_cond is 'device = yyyy'
        :param action_add: 添加新项
        :param action_delete: 删除缺失项
        :param action_mark_miss:  str, 如果mark_miss设置, 将不删除项, 转而设置col(mark_miss)=1
        :param func_skip:    func, 如果返回True, 则忽略更新
        :param param_query:    {condition=, key=}, query的参数
        :param param_add:
        :param param_update:
        :param param_delete:
        :param with_detail:  返回详细列表
        :param with_sqls:  返回详细sql
        :param timed:  添加更改时间，time_update
        :param last:  如果有last_<col>存在，自动更新
        :param use_df:  使用df方法
        :param df_to_str:  使用df时，两个df merge时, 类型不同会报错, 强制转成str
        :param block: 限制每次更新的项
        :param dryrun: 不实际执行
        :return:
        '''
        if not key:
            return False, 'no unique_key given.'
        if data_old is None:
            if param_query is None:
                param_query = {}
            if 'key' not in param_query:
                param_query['key'] = key
            is_ok, data_old = self.query(**param_query)
            assert is_ok, data_old
        if not param_add:
            param_add = {}
        if not param_update:
            param_update = {}
        if key and key not in param_update:
            param_update['key'] = key
        if not param_delete:
            param_delete = {}
        if isinstance(data_new, list):
            data_new = {x[key]: x for x in data_new}
        if isinstance(data_old, list):
            data_old = {x[key]: x for x in data_old}
        if use_df:
            d = self._update_compare_df(data_new, data_old, key, cols, func_skip, df_to_str)
        else:
            d = self._update_compare_dict(data_new, data_old, key, cols, func_skip, last=last)
        items_new, items_need_update, items_miss = d['items_new'], d['items_need_update'], d['items_miss']
        changed_cols = d['changed_cols']
        now = datetime_to_string()
        if items_new:
            if action_add and not dryrun:
                is_ok, msg = self.add_all(items_new, timed=timed, now=now, dryrun=dryrun, **param_add)
                assert is_ok, msg

        sqls = []
        if items_need_update:
            i = 0
            while True:
                if i > len(items_need_update):
                    break
                sub = items_need_update[i:i + block]
                is_ok, sql = self._update_case(changed_cols, sub, key, force_condition, timed=timed, now=now,
                                               dryrun=dryrun)
                assert is_ok, sql
                sqls.append(sql)
                i += block
        if items_miss:
            if action_mark_miss:
                cond = 'where {0} in ({1})'.format(key,
                                                   ','.join([add_quote(v[key]) for v in items_miss])) + \
                       'and {}'.format(force_condition) if force_condition else ''
                sql = 'update {0} set {1}=1 {2}'.format(self.model.__tablename__, action_mark_miss, cond)
                sqls.append(sql)
                if not dryrun:
                    r = self.session.execute(sql)
                    self.commit()
            elif action_delete:
                # self.session.query(self.model.id.in_((n,n,n))).all().delete()
                is_ok, sql = self.delete(data=items_miss, batch=True, key=key,
                                         force_condition=force_condition,
                                         **param_delete)
                assert is_ok, sql
                sqls.append(sql)
        if timed and 'time_check' in self.cols_name:
            cond = 'where {0} in ({1})'.format(key, ','.join(add_quote(list(data_new.keys())))) + (
                ' and {}'.format(force_condition) if force_condition else '')
            sql2 = 'update {0} set `{1}`="{2}" {3}'.format(self.model.__tablename__, 'time_check', now, cond)
            if not dryrun:
                r = self.session.execute(sql2)
                self.commit()

        l = [('input', len(data_new)),
             ('exist', len(data_old) - len(items_miss)),
             ('skiped', len(d['items_skip'])),
             ('nochange', len(data_old) - len(items_miss) - len(items_need_update)),
             ('changed', len(items_need_update)),
             ('only_changed_last', d['only_change_last']),
             ('dryrun', dryrun),
             ('updated', len(items_need_update)) if not dryrun else 0,
             ('action_add', action_add),
             ('new', len(items_new)),
             ('added', len(items_new) if action_add and not dryrun else 0),
             ('miss', len(items_miss)),
             ('action_delete', action_delete),
             ('deleted', len(items_miss) if action_delete and not dryrun else 0),
             ('action_mark_miss', action_mark_miss),
             ('marked', len(items_miss) if action_mark_miss and not dryrun else 0),
             ]
        r = OrderedDict()
        for k, v in l:
            r[k] = v
        if with_detail:
            for k in ['items_new', 'items_miss', 'items_need_update', 'items_skip']:
                r[k] = d[k]
        if with_sqls:
            r['sql'] = sqls

        return True, r

    def _update_case(self, cols, lst, key, force_condition, timed=True, now=None, time_col='time_update',
                     dryrun=False):
        '''
        :param cols:    cols should be modified
        :param lst:     [(data_new, data_old)]
        :param key:     index
        :param force_condition:
        :param timed:
        :param now:
        :param dryrun:
        :param time_col:
        :return:
        '''
        d = {c: [] for c in cols if c != key}
        l = []
        for i, x in enumerate(lst):
            new, old = x
            for c in cols:
                if c == key:
                    continue
                d[c].append('WHEN {0} THEN {1}'.format(self.add_quote_by_col_type(new[key], key),
                                                       self.add_quote_by_col_type(new[c], c)))
            l.append(self.add_quote_by_col_type(new[key], key))
        st = ',\n'.join([' `{0}` = case `{1}` '.format(k, key) + ' '.join(v) + ' END' for k, v in d.items()])
        cond = 'where {0} in ({1})'.format(key, ','.join(l)) + (
            ' and {}'.format(force_condition) if force_condition else '')
        sql = 'update {0}\n set {1} {2}'.format(self.model.__tablename__, st, cond)
        if not dryrun:
            r = self.session.execute(sql)
            self.commit()
        if timed and time_col in self.cols_name:
            sql1 = 'update {0} set `{1}`="{2}" {3}'.format(self.model.__tablename__, time_col, now, cond)
            if not dryrun:
                r = self.session.execute(sql1)
                self.commit()
        return True, sql

    @check_writable
    @catch_sql_exception
    def add(self, data, expandable=False, check=True):
        '''

        :param data:
        :param expandable:
        :param check: assure the data is not in db, called by update() with expandable set to False
        :return:
        '''
        if check:
            status, ex_obj = self.get_obj(data)
        else:
            ex_obj = None
        if ex_obj:
            if expandable:
                return self.update_one(data, ex_obj=ex_obj, expandable=False)
            else:
                return False, ex_obj.id
        data1 = {k: v for k, v in data.items() if k in self.cols_name}
        obj = self.model(**data1)
        self.session.add(obj)
        self.commit()
        return True, obj.id if hasattr(obj, 'id') else getattr(obj, self.cols_unique[0]) if self.cols_unique else ''

    @check_writable
    @catch_sql_exception
    def add_all(self, datas, block=1000, timed=True, now=None, time_col='time_insert', dryrun=False, **kwargs):
        '''
        :param datas:
        :param block:   分割块
        :param timed:   自动添加时间，time_insert
        :param now:   当前时间
        :param time_col:
        :param kwargs:
        :return:
        '''
        if timed and now and time_col in self.cols_name:
            for i, x in enumerate(datas):
                if time_col not in x.keys():
                    x[time_col] = now
                if 'time_update' in self.cols_name and 'time_update' not in x.keys():
                    x['time_update'] = now
        if block <= 0:
            block = 1000
        i = 0
        ids = []
        while True:
            _datas = datas[i * block:(i + 1) * block]
            if not _datas:
                break
            is_ok, _ids = self._add_all(_datas, **kwargs)
            assert is_ok, _ids
            ids.extend(_ids)
            i += 1
        return True, ids if kwargs.get('get_ids') else 'add all done.'

    @check_writable
    @catch_sql_exception
    def _add_all(self, datas, mode='save', get_ids=True):
        '''
        :param datas:
        :param mode: 'add' is operated one by one, return obj. 'save' is batch insert into db, no return by default
        :param get_ids: return ids of data
        :return:
        '''
        _datas = []
        for i, x in enumerate(datas):
            _datas.append({k: x.get(k) for k in self.cols_name})

        objs = [self.model(**_data) for _data in _datas]
        if mode == 'add':
            self.session.add_all(objs)
        elif mode == 'bulk':
            # python3 中会报 "string argument without an encoding"
            self.session.bulk_save_objects(objs)
        elif mode == 'save':

            cols = [k for k, v in datas[0].items() if k in self.cols_name]
            sql = 'insert into {} ({}) values {}'.format(
                self.model.__tablename__, ','.join(add_quote(cols, quote='`')),
                ','.join(['({})'.format(','.join(self.add_quote_by_col_type(x[k], k) for k in cols)) for x in datas]))
            # sql = sql.replace('None', 'null')
            self.session.execute(sql)
        self.commit()

        ids = []
        kw = None
        if not get_ids:
            return True, 'add all done.'
        if mode == 'save':
            for k in datas[0].keys():
                if datas[0][k] and k in self.cols_unique:
                    kw = k
                    break
            if not kw:
                return False, 'no unique key provided.'
            quote = False if isinstance(datas[0][kw], (int, float)) else True
            is_ok, resp = self.query(
                condition='{0} in ({1})'.format(kw,
                                                ','.join([('"{}"' if quote else '{}').format(x[k]) for x in datas])),
                key=kw)
            for i, x in enumerate(datas):
                id = resp.get(x[kw], {}).get('id')
                x['id'] = id
                ids.append(id)
        elif mode == 'add':
            for i, obj in enumerate(objs):
                datas[i]['id'] = obj.id
                ids.append(obj.id)
        return True, ids

    @check_writable
    @catch_sql_exception
    def delete(self, data=None, skip_miss=True, batch=True, key=None, condition=None, force_condition=None,
               dryrun=False):
        '''
        :param data:    can be dict/list of datas, data, obj
        :param skip_miss:   skip error that data not in db
        :param batch:   batch delete mode
        :param key:  set batch identifier in batch mode
        :param condition:  search condition when data is not given
        :param force_condition:  restrain the sql in the condition
        :param dryrun:
        :return:
        '''
        if isinstance(condition, str):
            if not re.match(r'where', condition.strip(), re.I):
                condition = 'where ' + condition
            sql = 'delete from {0} {1}'.format(self.model.__tablename__, condition)
            self.session.execute(sql)
            self.commit()
            return True, 'delete with condition done.'

        if batch:
            assert isinstance(data, (list, dict)), 'obj must be list or dict in batch delete mode.'
            for col in self.cols:
                if key == col.name:
                    pattern = '{0}' if isinstance(col.type, (INTEGER, FLOAT)) else '"{0}"'
                    break
            assert pattern, '{0} is not in model {1} when exec delete.'.format(key, self.model)
            if isinstance(data, list):
                values = ','.join(pattern.format(x[key]) for x in data)
            elif isinstance(data, dict):
                values = ','.join(pattern.format(v[key]) for k, v in data.items())
            sql = 'delete from `{0}` where `{1}` in ({2})'.format(self.model.__tablename__, key, values)
            sql += 'and {}'.format(force_condition) if force_condition else ''
            if not dryrun:
                self.session.execute(sql)
                # self.flush()
                self.commit()
            return True, sql
        else:
            if isinstance(data, dict):
                status, ex_obj = self.get_obj(data)
                if not ex_obj:
                    return skip_miss, 'data not exist.'
                obj = ex_obj
            else:
                obj = data
            if isinstance(obj, self.model):
                self.session.delete(obj)
                self.commit()
            return True, obj.id

    def commit(self):
        self.session.flush()
        self.session.commit()

    def close(self):
        self.session.close()

    @catch_sql_exception
    def search(self, q='', col=None, key=None, joint='-'):
        '''
        用self.query代替！！
        :param q: condition.
        :param key: can be str or list
        :param joint: if key is a list, the key of dict returned is concatenated by joint
        :return: return dict if key is set, else list
        '''
        if col is None:
            col_str = '*'
        elif isinstance(col, str):
            col_str = '`{}`'.format(col)
        elif isinstance(col, list):
            col_str = ','.join('`{}`'.format(x) for x in col)
        else:
            return False, 'wrong col name'
        sql = "select {0} from `{1}` {2} {3}".format(col_str, self.model.__tablename__, 'where' if q else '',
                                                     q)
        result = self.session.execute(sql)
        rows = result.fetchall()
        if not isinstance(rows, list):
            return False, 'search fail. SQL <{}>'.format(sql)

        if not key:
            datas = []
            if rows and len(rows) > 0:
                cols = rows[0]._parent.keys
                for i, row in enumerate(rows):
                    datas.append({col: row[j] for j, col in enumerate(cols)})
        else:
            datas = {}
            if rows and len(rows) > 0:
                cols = rows[0]._parent.keys
                keys = [key] if isinstance(key, str) else key
                keys_id = []
                for k in keys:
                    if k not in cols:
                        return False, '{0} is not in cols {1}'.format(k, ','.join(cols))
                    keys_id.append(cols.index(key))
                for i, row in enumerate(rows):
                    # k1 may be int, which can't be join
                    k1 = row[keys_id[0]] if len(keys_id) == 1 else joint.join(str(row[k]) for k in keys_id)
                    datas[k1] = {col: row[j] for j, col in enumerate(cols)}
        return True, datas

    @catch_sql_exception
    def _query1(self, sql, key=None, joint='-'):
        '''
        query1使用传统方式，可以在query2中返回同样的类型
        :param sql
        :param key: 可以是str或list
        :param joint: if key is a list, the key of dict returned is concatenated by joint
        :return: return dict if key is set, else list
        '''
        result = self.session.execute(sql)
        rows = result.fetchall()
        if not key:
            datas = []
            if rows and len(rows) > 0:
                cols = rows[0]._parent.keys
                for i, row in enumerate(rows):
                    datas.append({col: row[j] for j, col in enumerate(cols)})
        else:
            datas = {}
            if rows and len(rows) > 0:
                cols = rows[0]._parent.keys
                keys = [key] if isinstance(key, str) else key
                keys_id = []
                for k in keys:
                    if k not in cols:
                        return False, '{0} is not in cols {1}'.format(k, ','.join(cols))
                    keys_id.append(cols.index(key))
                for i, row in enumerate(rows):
                    k1 = joint.join(str(row[k]) for k in keys_id)
                    datas[k1] = {col: row[j] for j, col in enumerate(cols)}
        return True, datas

    def _query2(self, sql, ret_df=False, key=None, joint='-', apply_func=None, **kwargs):
        '''
        query2使用pandas方式
        :param sql:
        :param key:  str or list. different with 'index_col' in read_sql_query is this param will reserve the index in value.
        :param joint: if key is a list, the key of dict returned is concatenated by joint
        :param apply_func: apply函数，生成key
        :param kwargs: read_sql_query的kwargs
            index_col=['device_id]
            parse_dates=['finance_pay_date']
        :return:
        如果列col是int，但有None，该col会被返回成float
        '''
        df = pd.read_sql_query(sql, con=self.engine(), **kwargs)
        df = set_none(df)
        cols = df.columns.tolist()
        if ret_df:
            if apply_func:
                df.index = df.apply(apply_func, axis=1)
            if key:
                df.index = df.apply(lambda x: joint.join(x[k] for k in key) if isinstance(key, list) else x[key],
                                    axis=1)
            return True, df
        else:
            if not key:
                return True, df.to_dict(orient='record')
            else:
                if isinstance(key, list):
                    d = {joint.join([str(v[i]) for i in key]): v for k, v in df.to_dict(orient='index').items()}
                else:
                    d = {v[key]: v for k, v in df.to_dict(orient='index').items()}
                return True, d

    @catch_sql_exception
    def query(self, sql=None, table=None, condition='', cols=None, limit=None, orderby=None, **kwargs):
        '''
        使用_query2时, 如果sql中有%, 需要使用%%
        query2('table', '', ['a', ('b', 'b1')]) -> select a, b as b1 from table
        :param sql:   如果sql为空，则用后面的args组合sql
        :param table:   __tablename__
        :param condition:   <where> xxxx
        :param cols:    需要查询的columns， ['a', ('b', 'b1')]
        :param kwargs:
            ret_df  是否返回dataframe
        :return:
        '''
        if not sql:
            if not table:
                table = self.model.__tablename__
            if condition and not condition.strip().lower().split()[0] in ['where', 'left', 'right', 'join']:
                condition = 'where ' + condition
            if not cols:
                cols1 = '*'
                cols2 = None
            else:
                cols1 = ', '.join('{0} as {1}'.format(x[0], x[1]) if isinstance(x, tuple) else x for x in cols)
                cols2 = [x[1] if isinstance(x, tuple) else x for x in cols]

            sql = 'select {0} from {1} {2} {3} {4}'.format(cols1, table, condition,
                                                           'order by {}'.format(orderby) if orderby else '',
                                                           'limit {}'.format(limit) if limit else '')
        return self._query2(sql, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/10/30 10:57

import hashlib
import os
import re
import shutil
import socket
from pathlib import Path


def path_open(pathname, mode='r', **kwargs):
    '''
    defaultencoding和LC_ALL/LC_CTYPE/LANG变量有关，在pycharm中运行时，需要配置任一变量为C.UTF-8，
    如果zh_CN.UTF-8，有可能会改变term的语言
    '''
    sep = os.fsdecode(str(pathname)).split('/')[:-1]
    if re.search(r'(w|a)', mode):
        for i, name in enumerate(sep):
            if not name:
                continue
            _path = Path('/'.join(sep[:i + 1]))
            if not _path.is_dir():
                _path.mkdir()
    return open(os.fsencode(str(pathname)), mode, **kwargs)


def path_delete(pathname, force=False):
    '''
    删除空目录/文件
    :param pathname:
    :param force:   如果目录非空，也删除
    :return:
    '''
    pathname = Path(pathname)
    if pathname.exists():
        if pathname.is_file():
            pathname.unlink()
            return True, 'remove file {}.'.format(pathname)
        elif pathname.is_dir():
            if not [x for x in pathname.glob('*')]:
                pathname.rmdir()  # 空目录
                return True, 'remove empty dir {}.'.format(pathname)
            elif force:
                shutil.rmtree(str(pathname))
                return True, 'remove dir {} with files.'.format(pathname)
        return True, '{} is not file or dir.'.format(pathname)
    return True, '{} is not exist.'.format(pathname)


def path_copy(src, dst, overwrite=False, violent=False, **kwargs):
    '''
    :param src:
    :param dst:
    :param overwrite:   覆盖文件
    :param violent:     暴力，如果存在dst目录，删除后复制
    :param kwargs:
    :return:
    '''
    src, dst = Path(src), Path(dst)
    msg = '{} copy as {}.'.format(src, dst)
    if src.is_dir():
        if not dst.exists():
            shutil.copytree(src, dst)
            return True, msg
        else:
            if dst.is_dir():
                if violent:
                    shutil.rmtree(dst)
                    shutil.copytree(src, dst)
                    return True, msg
                else:
                    for x in src.glob('*'):
                        is_ok, _msg = path_copy(x, dst / x.name)
                        if not is_ok:
                            return is_ok, _msg
            else:
                return False, '{} is dir, but {} is a exist file.'.format(src, dst)
    else:
        if not dst.exists():
            shutil.copy2(src, dst)
            return True, msg
        else:
            if overwrite:
                shutil.copy2(src, dst)
                return True, '{} overwrite {}.'.format(src, dst)
            else:
                md5src = hashlib.md5(path_open(src)).hexdigest()
                md5dst = hashlib.md5(path_open(dst)).hexdigest()
                if md5src == md5dst:
                    return True, '{} is same as {}.'.format(dst, src)
                else:
                    return False, '{} is exist and different from {}.'.format(dst, src)


def check_port(ip, port):
    '''
    check port is available
    :param ip:
    :param port:
    :return:
    '''
    sk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sk.settimeout(2)
    try:
        sk.connect((ip, port))
        is_ok, msg = True, 'port {} is open.'.format(port)
    except Exception as e:
        is_ok, msg = False, 'port {} is close.'.format(port)
    sk.close()
    return is_ok, msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/26 11:38

# pip install concurrent-log-handler

import logging
import logging.config
from copy import deepcopy
from pathlib import Path

from ..os.info import get_pwd, get_caller


def make_logger(name=None, stem=None, log_dir='log', work_dir=None, console=True, write=True,
                focus_error=True, multi_process=True,
                level='INFO', **kwargs):
    '''
    :param name: logger name
    :param stem: file stem, default filename without type
    :param log_dir:
    :param work_dir:
    :param console:
    :param write:
    :param focus_error:     增加error日志
    :param multi_process:   支持多进程
    :param level:
    :param kwargs:
        {'handlers': {...}}
    :return:
    '''
    work_dir = work_dir or get_pwd()
    caller = get_caller()
    dpath = Path(work_dir) / log_dir
    dpath.mkdir(exist_ok=True)
    stem = stem or caller.stem
    name = name or caller.name

    # concurrent_log_handler.ConcurrentRotatingFileHandler 支持多进程，但不能做时间分割
    # logging.handlers.RotatingFileHandler 只支持多线程
    file_handler = {
        'level': 'DEBUG',
        'class': 'concurrent_log_handler.ConcurrentRotatingFileHandler' if multi_process else 'logging.handlers.RotatingFileHandler',
        'maxBytes': 1024 * 1024 * 10,
        'backupCount': 10,
        # If delay is true,
        # then file opening is deferred until the first call to emit().
        'delay': True,
        'filename': str(dpath / '{0}.log'.format(stem)),
        'formatter': 'verbose'
    }

    d = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'verbose': {
                'format': "[%(asctime)s] [%(filename)s:%(lineno)s] %(levelname)s %(message)s",
                'datefmt': "%Y-%m-%d %H:%M:%S"
            },
            'simple': {
                'format': '%(levelname)s %(message)s'
            },
        },
        'handlers': {
            'null': {
                'level': 'DEBUG',
                'class': 'logging.NullHandler',
            },
        },
    }

    if console:
        d['handlers']['console'] = {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        }
    if write:
        d['handlers']['file'] = file_handler
        if focus_error:
            file_handler_error = deepcopy(file_handler)
            file_handler_error.update({'level': 'WARN', 'filename': str(dpath / '{0}_error.log'.format(stem))})
            d['handlers']['file_error'] = file_handler_error
    if 'handlers' in kwargs:
        d['handlers'].update(kwargs['handlers'])

    d['loggers'] = {name: {'handlers': d['handlers'].keys(),
                           'level': level, }}

    logging.config.dictConfig(d)
    logger = logging.getLogger(name)
    return logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/20 14:57

import re


def comma_digit(s, tp=int):
    if str_is_number(s, tp, exp=True):
        return '{:,}'.format(int(s))
    return s


def str_is_number(s, tp=None, exp=True):
    '''

    :param s:
    :param tp: 指定int/float，None两者都行
    :param exp: 扩展，如果s是int/float，也可以返回true
    :return:
    '''
    if not isinstance(s, str):
        if exp and isinstance(s, int) and tp in (None, int):
            return True
        if exp and isinstance(s, float) and tp in (None, float):
            return True
        return False
    if tp == int:
        return re.match('^\d+\.{0,0}\d+$', s)
    elif tp == float:
        return re.match('^\d+\.{1,1}\d+$', s)
    else:
        return re.match('^\d+\.{,1}\d+$', s)


def str_is_email(s):
    if isinstance(s, str):
        return re.search('^(?P<id>.+)@(?P<domain>[\w\d]+\.([\w\d]+\.){0,}\w+$)', s)


def str2list(s):
    return [x.strip() for x in s.split(',') if x.strip()] if isinstance(s, str) else s


def bytes_decode(v, enconding='utf-8', errors='strict', **kwargs):
    '''
    convert v to spec type, default str, puresnmp use it.
    :param v:
    :param to_type:
    :param enconding:
    :param errors:
    :return:
    '''
    if isinstance(v, bytes):
        try:
            v = bytes.decode(v, encoding=enconding, errors=errors)
        except Exception as e:
            # if v is hex bytes
            v = v.hex()
    return change_type(v, **kwargs)


def change_type(v, to_type=None, default=None, strip=True):
    '''
    转换类型，如果有指定，但转换不了，返回default或原值，如果无指定，刚自动匹配
    :param v:
    :param _type:
    :param default: 默认返回，None则返回原值
    :param strip:
    :return:
    '''
    if to_type:
        try:
            if to_type in [str, 'str']:
                return str(v).strip() if strip else str(v)
            elif to_type in [int, 'int']:
                return int(v)
            elif to_type in [float, 'float']:
                return float(v)
        except Exception as e:
            return default if default is not None else v
    else:
        if str_is_number(v, int):
            return int(v)
        elif str_is_number(v, float):
            return float(v)
        return v


def replace(s, pats=None, default='', ret_with_pat=False, _any=False, flags=0, escape=False):
    '''
    :param s:
    :param pats:    替换特征, [(old, new)], 如果
    :param default:  如果传入的pats列表是str, 则用default的值进行替换
    :param ret_with_pat:  是否返回匹配列表
    :param _any:  匹配任意结束
    :param flags:  re flags
    :param escape:  传入的pats需要强制escape
    :return:
    '''
    if not pats:
        pats = [(r'[\r\n]+', '\n')]
    if isinstance(pats[0], str):
        pats = [(x, default) for x in pats]
    if escape:
        pats = [(re.escape(x1), x2) for x1, x2 in pats]
    _pats = '' if _any else []
    s1 = s
    for pat, rep in pats:
        s = re.sub(pat, rep, s, flags=flags)
        # 如果_any==True, 返回的_pats为str, 否则是list
        if s != s1:
            pat = pat.replace('\\', '')
            s1 = s
            if _any:
                _pats = pat
                break
            else:
                _pats.append(pat)
    if ret_with_pat:
        return s, _pats
    else:
        return s


def windows_filename(s, full=False, space=True):
    '''
    windows命名
    :param s:
    :param full: 用全角代替, 否则用_代替
    :param space: 是否替换空格
    :return:
    '''
    pats = [(r'\\', '＼'), (r'/', '／'), (r':', '：'), (r'\*', '＊'), (r'\?', '？'), (r'"', '＂'), (r'<', '＜'),
            (r'>', '＞'), (r'\|', '｜')]
    if space:
        pats.append((' ', '_'))
    if not full:
        pats = [(x[0], '_') for x in pats]
    return replace(s, pats=pats)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/4/29 22:43

__title__ = 'seakylib'
__author__ = 'Seaky'
__version__ = '0.0.2'
__copyright__ = 'Copyright (c) 2019 Seaky'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/8/14 12:05


import argparse
import json
import os
import pickle as pk
import re
import sys
import time
import traceback
from collections import OrderedDict
from functools import wraps
from pathlib import Path
from pprint import pprint

from ..func.log import make_logger
from ..os.info import get_pwd
from ..os.oper import path_open


def func_done(name=None):
    return 'func {} done.'.format(name if name else sys._getframe().f_back.f_code.co_name)


def count_time(f):
    @wraps(f)
    def wrap(*args, **kwargs):
        count_in_class = False
        if len(args) > 0:
            if hasattr(args[0], '__dict__') and not callable(args[0]):
                self = args[0]
                count_in_class = True
        start_time = time.time()
        result = f(*args, **kwargs)
        elapsed_time = round(time.time() - start_time, 2)
        ip = None
        if count_in_class:
            # 放到self.cache['timer']中
            if hasattr(self, 'cache') and isinstance(self.cache, dict):
                if not self.cache.get('timer'):
                    self.cache['timer'] = OrderedDict()
                self.cache['timer'][f.__name__] = elapsed_time
                if self.cache.get('debug') or (hasattr(self, 'debug') and getattr(self, 'debug')):
                    if hasattr(self, 'info'):
                        ip = self.info.get('snmp', {}).get('peername') or self.info.get('auth', {}).get('ip')
                    elif hasattr(self, 'ip'):
                        ip = self.ip
                    self.log.debug(
                        '{0:.2f}s {1}{2}.'.format(elapsed_time, f.__name__, ' on {}'.format(ip) if ip else ''))
        else:
            if True or kwargs.get('debug'):
                print('{0:.2f}s {1}{2}.'.format(elapsed_time, f.__name__, ' on {}'.format(ip) if ip else ''))
        return result

    return wrap


def is_instance(obj):
    '''对象是一个实例'''
    if hasattr(obj, '__dict__') and not callable(obj):
        return True


def catch_exception(ignore=False, retry=0):
    '''
    :param ignore: 不记录错误
    :param retry: 重试次数
    :return:
    如果 kwargs中含有expose_exception，忽略catch
    '''

    def deco(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            if 'expose_exception' in kwargs:
                expose_exception = kwargs.pop('expose_exception')
                if expose_exception:
                    return f(*args, **kwargs)
            in_class = False
            if len(args) > 0:
                if is_instance(args[0]):
                    self = args[0]
                    in_class = True
            i = 0
            while i <= retry:
                try:
                    excep = False
                    result = f(*args, **kwargs)
                    break
                except AssertionError as e:
                    excep, message, trace = True, '{}'.format(e), '{}'.format(traceback.format_exc())
                except Exception as e:
                    excep, message, trace = True, '{}'.format(e), '{}'.format(traceback.format_exc())
                # print('--{}--'.format(i), retry, self.ip, f, trace)
                i += 1
            if not excep:
                if in_class:
                    self.cache['exception']['critical'] = False
                return result
            else:
                verbose = kwargs.get('verbose')
                if in_class:
                    if 'errors' not in self.cache['exception']:
                        self.cache['exception']['error'] = []
                    self.cache['exception']['critical'] = not ignore
                    self.cache['exception']['error'].append(
                        {'func': f.__name__, 'error_message': '{}'.format(message),
                         'error_traceback': '{}'.format(trace)})
                    if hasattr(self, 'verbose') and getattr(self, 'verbose'):
                        verbose = verbose or self.verbose
                    if self.log and not ignore:
                        self.log.debug(trace)
                return False, trace if verbose else message

        return wrap

    return deco


def run_func(obj, **kwargs):
    '''
    运行传入的函数，func/(func, (...))/(func, {...}), kwargs也会被传递给func
    :param obj:
    :return:
    '''
    if hasattr(obj, '__call__'):
        return obj()
    elif isinstance(obj, tuple):
        if len(obj) == 2:
            func, args = obj
            if isinstance(args, tuple):
                return func(*args, **kwargs)
            elif isinstance(args, dict):
                return func(**args, **kwargs)
        else:
            func, *args = obj
            return func(*args, **kwargs)
    return False, '{} can not be called'.format(obj)


@catch_exception()
def run_functions(*funcs, message=None, order='and', watchdog=None, **kwargs):
    '''
    funcs见run_func
    如果是and，所有func返回True，最后返回True
    如果是or，任意func返回True即返回True，否则返回False
    watchdog: 额外判断函数
    '''

    for func in funcs:
        is_ok, _message = run_func(func)
        assert is_ok, _message
        x, y = run_func(watchdog)
        if not x:
            is_ok, _message = x, '{} ({})'.format(_message, y)
        if order == 'and':
            if not is_ok:
                return is_ok, _message
        elif order == 'or':
            if is_ok:
                return is_ok, _message
    # 如果是and，跳出循环表明都执行成功；如果是or，跳出循环表明都执行失败
    if order == 'and':
        return True, message if message else 'run_fucs successfully.'
    elif order == 'or':
        return False, message if message else 'run_fucs fail.'


class MyClass:
    def __init__(self, *args, **kwargs):
        '''
        :param args:
        :param kwargs:
            path_output
            path_temp
            log:    customer log
            log_params
            show_tick  显示count_time的值
            log_params  make_logger的参数
            quite
            debug   logging.debug
        '''
        self.kwargs = kwargs
        self.pwd = get_pwd()
        self.path_output = self.pwd / Path(kwargs.get('path_output', 'output'))
        self.path_temp = self.pwd / Path(kwargs.get('path_temp', 'temp'))
        # 默认log是打印到console
        log_params = kwargs.get('log_params', {})
        if kwargs.get('debug'):
            log_params.update({'level': 'DEBUG'})
        # self.log = kwargs.get('log', make_logger(self.__class__.__name__, **log_params))
        # 两个类建立默认log时，前者会没有输出？
        self.log = kwargs['log'] if 'log' in kwargs else make_logger(self.__class__.__name__, **log_params)
        self.quite = kwargs.get('quite')
        # cache存放过程数据, 最好不要放入无法dump的数据
        self.cache = {'error': [], 'timer': OrderedDict(), 'exception': {'critical': False}, 'message': ''}
        self.verbose = kwargs.get('verbose')

    def show_verbose(self, *obj):
        if self.verbose:
            for o in obj:
                pprint(o) if isinstance(o, (list, dict)) else print(o)

    def log_error(self, s, *kwargs):
        '''在这里记录错误'''
        self.cache['error'].append(s)
        self.log.error(s, *kwargs)


def load_data(filename, store_dir='temp', work_dir=None, mode='json'):
    work_dir = work_dir or get_pwd()
    dpath = Path(work_dir) / store_dir
    fpath = dpath / os.fsdecode(filename)
    if os.path.exists(str(fpath)):
        try:
            if mode == 'pickle':
                return True, pk.load(path_open(str(fpath), 'rb'))
            elif mode == 'json':
                return True, json.load(path_open(str(fpath), 'r'))
        except Exception as e:
            return False, ''


def dump_data(obj, filename, store_dir='temp', work_dir=None, mode='json'):
    work_dir = work_dir or get_pwd()
    dpath = Path(work_dir) / store_dir
    fpath = dpath / os.fsdecode(str(filename))
    dpath.mkdir(exist_ok=True)
    if mode == 'pickle':
        pk.dump(obj, path_open(str(fpath), 'wb'))
    elif mode == 'json':
        json.dump(obj, path_open(str(fpath), 'w'), indent='  ')
    return True, ''


# def a(func):
#     if func
#
# # save load需要编写
# def sl(obj, filename, func=None, save=True, load=True, store_dir='temp', work_dir=None, mode='json'):
#     work_dir = work_dir or get_pwd()
#     dpath = Path(work_dir) / store_dir
#     fpath = dpath / os.fsdecode(filename)
#     if load and fpath.exists():
#         data = json.load(path_open(fpath))
#         return True, data
#     self.kwargs['os'] = self.data_in_db['device']['os']
#     is_ok, result = self.inspect(method='snmp' if self.os == 'comware' else 'cli')
#     assert is_ok, result
#     self.cache['inspect_done'] = True
#     if self.kwargs.get('save') and is_ok:
#         json.dump(self.info, path_open(path_data, 'w'))
#     self.data_new = self.info
#     return is_ok, result


def func_result(save=False, load=False, ident_key=None, mark=None, func_dump=None, func_load=None):
    '''
    保存成 <store_dir>/[mark_]<func_name>[_ident_key]，一些特殊的对象无法保存
    :param save:
    :param load:
    :param ident_key:
    :param path:
    :param mark:
    :param func_dump:
    :param func_load:
    :return:
    '''

    def deco(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            filename = f.__name__
            if ident_key and ident_key in kwargs:
                filename += '_{}={}'.format(ident_key, kwargs[ident_key])
            if mark:
                filename = '{}_{}'.format(mark, filename)
            else:
                if 'mark' in kwargs:
                    filename = '{}_{}'.format(kwargs['mark'], filename)
            filename += '.result'
            filename = filename.replace('/', '-')
            in_class = False
            if len(args) > 0:
                if is_instance(args[0]):
                    self = args[0]
                    in_class = True
                    if hasattr(self, 'path_temp'):
                        filename = Path(self.path_temp) / filename
            if load and filename.is_file():
                r, d = load_data(filename)
                if r:
                    return func_load(d['result']) if func_load else d['result']
                else:
                    return
            result = f(*args, **kwargs)
            if save:
                r, d = dump_data({'result': func_dump(result) if func_dump else result,
                                  'func': f.__name__, 'args': str(args), 'kwargs': str(kwargs)}, filename)
            return result

        return wrap

    return deco


class ArgParseClass:
    def __init__(self, description='', group=None):
        '''
        :param description:
        :param group: add_all()中优先于__init__
        '''
        self.parser = argparse.ArgumentParser(description=description)
        self.group = group
        self.groups = {}
        if group:
            self.groups[group] = self.parser.add_argument_group(title=group)
        self.opts = []
        # require 强制，hidden 隐藏arg, forbidden 禁止某些arg。
        # 这些列表是为继承时检查，都不需要加--
        self.require = []
        self.hidden = []
        self.forbidden = []

    def add_all(self):
        # 上层可以按需重写 add_all
        pass

    def add_base(self, group='Base'):
        self.add('--verbose', action='store_true', help='verbose mode', group=group)

    def add_notify(self, group='Notify'):
        self.add('--yx', action='store_true', help='yixin', group=group)
        self.add('--popo', action='store_true', help='popo', group=group)
        self.add('--sms', action='store_true', help='sms', group=group)

    def add(self, *args, **kwargs):
        opt = args[0]
        # 隐藏参数
        if re.sub('^-+', '', opt) in self.hidden:
            return
        if opt in self.opts:
            return
        title = kwargs.get('group') or self.group
        if title:
            if title not in self.groups:
                self.groups[title] = self.parser.add_argument_group(title=title)
            parser = self.groups[title]
        else:
            parser = self.parser
        if 'group' in kwargs:
            del kwargs['group']
        self.opts.append(opt)
        parser.add_argument(*args, **kwargs)

    def parse(self):
        self.args = self.parser.parse_args()
        return self.args

    def dict(self):
        return self.parse().__dict__

    def check_args(self):
        '''
        检查require和forbidden列表
        :return:
        '''
        cmd = ' '.join(sys.argv)
        if '--help' in cmd:
            return
        for x in self.require:
            if not re.search(' -+{}'.format(x), cmd):
                self.parser.print_help()
                print('\narg "--{}" is required !!!\n'.format(x))
                exit(1)
        for x in sys.argv[1:]:
            if x.startswith('--') and re.sub('^-+', '', x) in self.forbidden:
                self.parser.print_help()
                print('\narg "{}" is forbidden !!!\n'.format(x))
                exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/6/25 10:00

import json
import random
import re
import warnings
from functools import wraps
from pathlib import Path
from urllib import parse

import requests
from bs4 import BeautifulSoup

from ..func.base import MyClass

warnings.filterwarnings("ignore")

UA = {'ie': 'Mozilla/5.0 (MSIE 10.0; Windows NT 6.1; Trident/5.0)',
      'chrome': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/65.0.3325.181 Safari/537.36',
      'firefox': 'Mozilla/5.0 (Windows NT 6.1; rv:2.0.1) Gecko/20100101 Firefox/4.0.1',
      'iphone': 'Mozilla/5.0 (iPhone; U; CPU iPhone OS 4_3_3 like Mac OS X; en-us) AppleWebKit/533.17.9 (KHTML, like Gecko) Version/5.0.2 Mobile/8J2 Safari/6533.18.5',
      'android': 'Mozilla/5.0 (Linux; U; Android 2.3.7; en-us; Nexus One Build/FRF91) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1',
      'wx': 'Mozilla/5.0 (iPhone; CPU iPhone OS 7_0_4 like Mac OS X) AppleWebKit/537.51.1 (KHTML, like Gecko) Mobile/11B554a MicroMessenger/6.2.1'}


class Http(MyClass):
    def __init__(self, random_ua=False, *args, **kwargs):
        '''
        :param random_ua:   随机UA
        :param args:
        :param kwargs:
        '''
        MyClass.__init__(self, *args, **kwargs)
        self.kwargs = kwargs
        self.session = requests.session()
        if random_ua:
            self.session.headers.update({'User-Agent': UA[kwargs.get('ua', 'chrome')]})
        else:
            tag, ua = random.choice(list(UA.items()))
            self.session.headers.update({'User-Agent': ua})
        self.proxies = kwargs.get('proxies')
        self.ssl_verify = kwargs.get('ssl_verify', False)
        self.status = {}  # for some vars preserve
        self.url_root = kwargs.get('url_root')

    def __getattr__(self, item):
        return self.__dict__.get(item, self.kwargs.get(item))

    def fetch(self, url, ret_bs=True, ret_raw=False, method='GET', tries=3, **kwargs):
        d = {'verify': False, 'timeout': 40, 'proxies': self.proxies}
        d.update(kwargs)
        if not re.search(r'^http', url, re.I) and self.url_root:
            url = re.sub('/*$', '', self.url_root) + '/' + re.sub('^/*', '', url)
        flag = False
        for i in range(tries):
            try:
                _raw = self.session.post(url, **d) if method == 'POST' else self.session.get(url, **d)
                flag = True
                break
            except Exception as e:
                error = e
        if not flag:
            raise Exception('fetch {} error({}). {}'.format(url, error, d))
        if ret_raw:
            self.fetch_after(_raw, None, None)
            return _raw
        _raw = _raw.content
        m = re.search('charset=\W*(?P<charset>\w+)', _raw[:200].decode(errors='ignore'))
        charset = m.groupdict().get('charset', 'utf-8') if m else 'utf-8'
        if charset == 'gb2312':
            charset = 'cp936'
        _content = _raw.decode(encoding=charset)
        bs = BeautifulSoup(_content, features=self.kwargs.get('features', 'html.parser'))
        self.fetch_after(_raw, _content, bs)
        return bs if ret_bs else _content

    def fetch_after(self, *args):
        '''
        第次交互后，可能需要执行的动作，如获取token
        :param args:
        :return:
        '''
        return True

    def get(self, *args, **kwargs):
        return self.fetch(method='GET', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.fetch(method='POST', *args, **kwargs)

    def save_stat(self, fn=None):
        fn = fn or 'status.json'
        json.dump(
            {'cookies': self.session.cookies.get_dict(), 'status': self.status},
            open(str(fn), 'w'), sort_keys=True, indent=True)

    def load_stat(self, fn=None):
        fn = fn or 'status.json'
        if Path(fn).exists():
            d = json.load(open(str(fn)))
            self.session.cookies.update(d['cookies'])
            self.status.update(d['status'])
            return True

    def login(self, load=True, save=True):
        '''登陆'''
        if load and self.load_stat():
            if self.login_verify():
                self.cache['login'] = True
                return True
        if self.login_action():
            save and self.save_stat()
            self.cache['login'] = True
            return True

    def login_action(self):
        '''
        实际登陆过程，重写
        :return:
        '''
        pass

    def login_verify(self):
        '''
        self.login()时，验证load数据
        :return:
        '''
        pass

    def job(self):
        pass

    def run(self):
        self.job()


def login_check(f):
    @wraps(f)
    def wrap(self, *args, **kwargs):
        if self.cache.get('login'):
            return f(self, *args, **kwargs)
        else:
            return False

    return wrap


def url2list(url):
    # 转dict需要注意相同的key会被覆盖
    return parse.parse_qsl(url)


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/6/18 14:01

import json
from socketserver import StreamRequestHandler, ForkingTCPServer

from ..func.base import MyClass

ForkingTCPServer.allow_reuse_address = True


class MyRequestHandle(StreamRequestHandler):
    # log = make_logger('Server', write=True)
    log = None
    encoding = 'utf-8'

    def handle(self):
        self.log.info('Got connection from {}'.format(self.client_address))
        for line in self.rfile:
            line = line.decode(self.encoding)
            self.log.info('receive from {} {}'.format(self.client_address, repr(line)))
            self.job(line)

    def job(self, recv):
        self.reply('receive ' + recv)

    def reply(self, message):
        if isinstance(message, (list, dict)):
            message = json.dumps(message).encode(self.encoding)
        if isinstance(message, bytes):
            message = message.decode(self.encoding)
        self.wfile.write(message.encode(self.encoding))
        self.log.info('send to {} {}'.format(self.client_address, repr(message)))


class MyTCPServer(MyClass):
    def __init__(self, port, handle, *args, **kwargs):
        MyClass.__init__(self, *args, **kwargs)
        self.port = port
        self.handle = handle
        self.handle.log = self.log

    def run(self):
        serv = ForkingTCPServer(('', self.port), self.handle)
        self.log.info('start server on port {}'.format(self.port))
        serv.serve_forever()


if __name__ == '__main__':
    ms = MyTCPServer(port=20000, handle=MyRequestHandle, log_params={'write': True, 'console': False})
    ms.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/6/25 16:07


import setuptools

import seakylib

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=seakylib.__title__,
    version=seakylib.__version__,
    author=seakylib.__author__,
    author_email='seaky.cn@gmail.com',
    description='Seaky\'s private lib',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sseaky',
    packages=setuptools.find_packages(),
    install_requires=['bs4==0.0.1','IPy==1.0','netmiko==2.4.2','numpy==1.17.3','pandas==0.24.2','psutil==5.6.4','puresnmp==1.6.3','PyMySQL==0.9.3','requests==2.22.0','SocksiPy-branch==1.1','sqlacodegen==2.1.0','SQLAlchemy==1.3.10','sshtunnel==0.1.5'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Seaky
# @Date:   2019/10/10 10:31

import socket

# pip install SocksiPy-branch
import socks
from sshtunnel import SSHTunnelForwarder


def set_socks_proxy(addr=None, port=None, rdns=True, username=None, password=None,
                    proxytype=socks.PROXY_TYPE_SOCKS5):
    '''
    使用已有socks5
    proxytype -    The type of the proxy to be used. Three types
            are supported: PROXY_TYPE_SOCKS4 (including socks4a),
            PROXY_TYPE_SOCKS5 and PROXY_TYPE_HTTP
    addr -        The address of the server (IP or DNS).
    port -        The port of the server. Defaults to 1080 for SOCKS
            servers and 8080 for HTTP proxy servers.
    rdns -        Should DNS queries be preformed on the remote side
            (rather than the local side). The default is True.
            Note: This has no effect with SOCKS4 servers.
    username -    Username to authenticate with to the server.
            The default is no authentication.
    password -    Password to authenticate with to the server.
            Only relevant when username is also provided.
    '''
    defaultproxy = {'proxytype': proxytype, 'addr': addr, 'port': port, 'rdns': rdns,
                    'username': username, 'password': password}
    socks.setdefaultproxy(**defaultproxy)
    socket.socket = socks.socksocket


def unset_socks_proxy():
    socks.setdefaultproxy(None)
    socket.socket = socks.socksocket


class Tunnel:
    def __init__(self, ssh_address_or_host, ssh_username, ssh_password=None, ssh_pkey=None,
                 ssh_private_key_password=None, **kwargs):
        self.kwargs = kwargs
        self.kwargs.update({'ssh_address_or_host': ssh_address_or_host,
                            'ssh_username': ssh_username, 'ssh_password': ssh_password,
                            'ssh_pkey': ssh_pkey, 'ssh_private_key_password': ssh_private_key_password})

    def setup(self, remote_bind_address, local_bind_address=None):
        '''
        :param remote_bind_address: (x.x.x.x, port)
        :param local_bind_address: 随机分配, server.local_bind_port
        :return:
        '''
        server = SSHTunnelForwarder(
            remote_bind_address=remote_bind_address,
            local_bind_address=local_bind_address,
            **self.kwargs
        )
        self.server = server
        return server
